'''
Features for use in simulation tasks
'''
import numpy as np
from riglib.bmi import clda, extractor, sim_neurons, kfdecoder, train
from riglib.bmi.sim_neurons import KalmanEncoder

class FakeHDF(object):
    def __init__(self):
        self.msgs = []

    def sendMsg(self, msg):
        self.msgs.append(msg)


class SimHDF(object):
    '''
    An interface-compatbile HDF for simulations which do not require saving an
    HDF file
    '''
    def __init__(self, *args, **kwargs):
        '''
        Constructor for SimHDF feature

        Parameters
        ----------
        args, kwargs: None necessary

        Returns
        -------
        SimHDF instance
        '''
        from collections import defaultdict
        self.data = defaultdict(list)
        self.task_data_hist = []
        self.msgs = []        
        self.hdf = FakeHDF()

        super(SimHDF, self).__init__(*args, **kwargs)

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' creates a fake task data variable so that 
        code expecting SaveHDF runs smoothly.
        '''
        super(SimHDF, self).init()
        self.dtype = np.dtype(self.dtype)
        self.task_data = np.zeros((1,), dtype=self.dtype)

    def sendMsg(self, msg):
        '''
        Simulate the "message" table of the HDF file associated with each source

        Parameters
        ----------
        msg: string
            Message to store

        Returns
        -------
        None
        '''
        self.msgs.append((msg, -1))

    def _cycle(self):
        super(SimHDF, self)._cycle()
        self.task_data_hist.append(self.task_data.copy())


class SimClock(object):
    def __init__(self, *args, **kwargs):
        super(SimClock, self).__init__(*args, **kwargs)

    def tick(self, *args, **kwargs):
        pass

    def get_fps(self):
        return 0.

class SimClockTick(object):
    '''
    Summary: A simulation pygame.clock to use in simulations that inherit from experiment.Experiment, to overwrite
    the pygame.clock.tick in the ._cycle function ( self.clock.tick(self.fps) )
    '''
    def __init__(self, *args, **kwargs):
        '''
        Summary: Constructor for SimClock
        Input param: *args:
        Input param: **kwargs:
        Output param: 
        '''
        super(SimClockTick, self).__init__(*args, **kwargs)

    def init(self, *args, **kwargs):
        self.clock = SimClock()
        super(SimClockTick, self).init(*args, **kwargs)


class SimTime(object):
    '''
    An accelerator so that simulations can run faster than real time (the task doesn't try to 'sleep' between loop iterations)
    '''
    def __init__(self, *args, **kwargs):
        '''
        Constructor for SimTime

        Parameters
        ----------
        *args, **kwargs: passed to parent (super) constructor

        Returns
        -------
        None
        '''
        super(SimTime, self).__init__(*args, **kwargs)
        self.start_time = 0.

    def get_time(self):
        '''
        Simulates time based on Delta*cycle_count, where the update_rate is specified as an instance attribute
        '''
        try:
            if not (self.cycle_count % (60*10)):
                print(self.cycle_count/(60*10.))
            return self.cycle_count * self.update_rate

        except:
            # loop_counter has not been initialized yet, return 0
            return 0

    @property 
    def update_rate(self):
        '''
        Attribute for update rate of task. Using @property in case any future modifications
        decide to change fps on initialization
        '''
        return 1./60


class SimNeuralEnc(object):
    def __init__(self, *args, **kwargs):
        if not hasattr(self, 'fb_ctrl'):
            self.fb_ctrl = kwargs.pop('fb_ctrl', None)
        if not hasattr(self, 'ssm'):
            self.ssm = kwargs.pop('ssm', None)
        super(SimNeuralEnc, self).__init__(*args, **kwargs)

    def init(self,):
        self._init_neural_encoder()
        self.wait_time = 0
        self.pause = False
        print('neural encoder init function ', self)
        super(SimNeuralEnc, self).init()

    def change_enc_ssm(self, new_ssm):
        self.encoder_old = self.encoder
        C_old = self.encoder_old.C.copy()
        Q = self.encoder_old.Q.copy()
        old_ssm = np.array(self.encoder_old.ssm.state_names).copy()

        old_to_new_map = np.squeeze(np.array([np.nonzero(np.array(new_ssm.state_names)==st)[0] for st in old_ssm]))

        self.ssm = new_ssm
        self._init_neural_encoder()

        self.encoder.C[:,old_to_new_map] =  C_old
        self.encoder.Q = Q

    def _init_neural_encoder(self):
        raise NotImplementedError


class SimKalmanEnc(SimNeuralEnc):
    def _init_neural_encoder(self):
        n_features = 50
        self.encoder = KalmanEncoder(self.ssm, n_features)

    def create_feature_extractor(self):
        '''
        Create the feature extractor object
        '''
        self.extractor = extractor.SimDirectObsExtractor(self.fb_ctrl, self.encoder, 
            n_subbins=self.decoder.n_subbins, units=self.decoder.units, task=self)
        self._add_feature_extractor_dtype()

    def create_feature_accumulator(self):
        '''
        Instantiate the feature accumulator used to implement rate matching between the Decoder and the task,
        e.g. using a 10 Hz KFDecoder in a 60 Hz task
        '''
        from riglib.bmi import accumulator
        feature_shape = [self.decoder.n_features, 1]
        feature_dtype = np.float64
        acc_len = int(self.decoder.binlen / self.update_rate)
        acc_len = max(1, acc_len) 
        self.feature_accumulator = accumulator.RectWindowSpikeRateEstimator(acc_len, feature_shape, feature_dtype)


class SimCosineTunedEnc(SimNeuralEnc):
    
    def _init_neural_encoder(self):
        ## Simulation neural encoder
        from riglib.bmi.sim_neurons import GenericCosEnc#CLDASimCosEnc
        print('SimCosineTunedEnc SSM:', self.ssm)
        self.encoder = GenericCosEnc(self.sim_C, self.ssm, return_ts=True, DT=0.1, call_ds_rate=6)
        
    def create_feature_extractor(self):
        '''
        Create the feature extractor object
        '''
        self.extractor = extractor.SimBinnedSpikeCountsExtractor(self.fb_ctrl, self.encoder, 
            n_subbins=self.decoder.n_subbins, units=self.decoder.units, task=self)
        self._add_feature_extractor_dtype()

class SimNormCosineTunedEnc(SimNeuralEnc):

    def _init_neural_encoder(self):
        from riglib.bmi.sim_neurons import NormalizedCosEnc
        self.encoder = NormalizedCosEnc(self.plant.endpt_bounds, self.sim_C, self.ssm, spike=True, return_ts=False, 
            DT=self.update_rate, tick=self.update_rate, gain=self.fov)
    
    def create_feature_extractor(self):
        '''
        Create the feature extractor object
        '''
        self.extractor = extractor.SimDirectObsExtractor(self.fb_ctrl, self.encoder, 
            n_subbins=self.decoder.n_subbins, units=self.decoder.units, task=self)
        self._add_feature_extractor_dtype()

class SimLFPCosineTunedEnc(SimNeuralEnc):

    bands = [(51, 100)]

    def _init_neural_encoder(self):
        from riglib.bmi.sim_neurons import NormalizedCosEnc
        self.encoder = NormalizedCosEnc(self.plant.endpt_bounds, self.sim_C, self.ssm, spike=False, return_ts=False, 
            DT=self.update_rate, tick=self.update_rate, n_bands=len(self.bands), gain=self.fov)
    
    def create_feature_extractor(self):
        '''
        Create the feature extractor object
        '''
        self.extractor = extractor.SimPowerExtractor(self.fb_ctrl, self.encoder, 
            channels=self.decoder.channels, bands=self.bands, task=self)
        self._add_feature_extractor_dtype()

class SimFAEnc(SimCosineTunedEnc):
    def __init__(self, *args, **kwargs):
        self.FACosEnc_kwargs = kwargs.pop('SimFAEnc_kwargs', dict())
        super(SimFAEnc, self).__init__(*args, **kwargs)

    def _init_neural_encoder(self):
        if 'encoder_fname' in self.FACosEnc_kwargs:
            import pickle
            self.encoder = pickle.load(open(self.FACosEnc_kwargs['encoder_fname']))
            print('using saved encoder;')
            
            if 'wt_sources' in self.FACosEnc_kwargs:
                self.encoder.wt_sources = self.FACosEnc_kwargs['wt_sources']
                print('setting new weights: ', self.encoder.wt_sources)
        
        else:
            from riglib.bmi.sim_neurons import FACosEnc
            self.encoder = FACosEnc(self.sim_C, self.ssm, return_ts=True, DT=0.1, call_ds_rate=6, **self.FACosEnc_kwargs)


class SimCosineTunedPointProc(SimNeuralEnc):
    def _init_neural_encoder(self):
        # from riglib.bmi import sim_neurons
        # decoder_fname = '/storage/decoders/grom20150102_14_BPPF01021655.pkl'
        # decoder = pickle.load(open(decoder_fname))

        # beta = np.array(decoder.filt.C)
        # dt = decoder.filt.dt 
        # self.encoder = sim_neurons.CLDASimPointProcessEnsemble(beta, dt)


        # 2) create a fake beta
        const = -1.6
        alpha = 0.10
        n_cells = 30
        angles = np.linspace(0, 2*np.pi, n_cells)
        beta = np.zeros([n_cells, 7])
        beta[:,-1] = -1.6
        beta[:,3] = np.cos(angles)*alpha
        beta[:,5] = np.sin(angles)*alpha
        beta = np.vstack([np.cos(angles)*alpha, np.sin(angles)*alpha, np.ones_like(angles)*const]).T

        beta_full = np.zeros([n_cells, 7])
        beta_full[:,[3,5,6]] = beta
        # dec = train._train_PPFDecoder_sim_known_beta(beta_full, encoder.units, dt=dt)

        # create the encoder
        dt = 1./180
        self.encoder = sim_neurons.CLDASimPointProcessEnsemble(beta_full, dt)
        self.beta_full = beta_full

    def create_feature_extractor(self):
        '''
        Create the feature extractor object
        '''
        self.extractor = extractor.SimBinnedSpikeCountsExtractor(self.fb_ctrl, self.encoder, 
            n_subbins=self.decoder.n_subbins, units=self.decoder.units, task=self)
        self._add_feature_extractor_dtype()        

#############################
##### Simulation Decoders
#############################


class SimKFDecoder(object):
    '''
    General class for all SimDFDecoder classes
    '''
    def __init__(self, *args, **kwargs):
        super(SimKFDecoder, self).__init__(*args, **kwargs)

    def change_dec_ssm(self):
        decoder_old = self.decoder_old
        ssm_old = decoder_old.ssm
        
        #Map new stuff
        old_to_new_map = np.squeeze(np.array([np.nonzero(np.array(self.ssm.state_names)==st)[0] for st in ssm_old.state_names]))

        #Update C: 
        C = self.decoder.kf.C.copy()
        C[:,old_to_new_map] = decoder_old.filt.C

        #Keep current dtype format: 
        dtype_new = self.dtype[:]
       
        #New KF: 
        kf = kfdecoder.KalmanFilter(self.decoder.kf.A, self.decoder.kf.W, C, self.decoder.kf.Q, is_stochastic=self.ssm.is_stochastic)
        #New decoder: 
        self.decoder = kfdecoder.KFDecoder(kf, self.decoder.units, self.ssm, mFR=self.decoder.mFR, sdFR=self.decoder.sdFR, binlen=self.decoder.binlen, tslice=self.decoder.tslice)

        #Replace dtype as it was before running KF functions
        self.dtype = dtype_new

        # Compute sufficient stats for C and Q matrices (used for RML CLDA)
        n_features, n_states = C.shape
        R = np.mat(np.zeros([n_states, n_states]))
        S = np.mat(np.zeros([n_features, n_states]))
        R_small, S_small, T, ESS = clda.KFRML.compute_suff_stats(self.init_kin_features[self.ssm.train_inds, :], self.init_neural_features)

        R[np.ix_(self.ssm.drives_obs_inds, self.ssm.drives_obs_inds)] = R_small
        S[:,self.ssm.drives_obs_inds] = S_small

        self.decoder.filt.R = R
        self.decoder.filt.S = S
        self.decoder.filt.T = T
        self.decoder.filt.ESS = ESS
        self.decoder.n_features = n_features


    def load_decoder(self):
        ''' Overwritten in child classes '''
        pass


class SimKFDecoderSup(SimKFDecoder):
    '''
    Construct a KFDecoder based on encoder output in response to states simulated according to the state space model's process noise
    '''
    def load_decoder(self):
        '''
        Instantiate the neural encoder and "train" the decoder
        '''
        
        if hasattr(self, 'decoder'):
            print('Already have a decoder!')
        else:
            print("Creating simulation decoder..")
            print(self.encoder, type(self.encoder))
            n_samples = 2000
            units = self.encoder.get_units()
            n_units = len(units)
            print('units: ', n_units)

            # draw samples from the W distribution
            ssm = self.ssm
            A, _, W = ssm.get_ssm_matrices()
            mean = np.zeros(A.shape[0])
            mean[-1] = 1
            state_samples = np.random.multivariate_normal(mean, W, n_samples)

            spike_counts = np.zeros([n_units, n_samples])
            self.encoder.call_ds_rate = 1
            for k in range(n_samples):
                spike_counts[:,k] = np.array(self.encoder(state_samples[k], mode='counts')).ravel()

            kin = state_samples.T
            zscore = False
            if hasattr(self, 'clda_adapt_mFR_stats'):
                if self.clda_adapt_mFR_stats:
                    zscore = True
            print(' zscore decoder ? : ', zscore)
            self.decoder = train.train_KFDecoder_abstract(ssm, kin, spike_counts, units, 0.1, zscore=zscore)
            self.encoder.call_ds_rate = 6

            self.init_neural_features = spike_counts
            self.init_kin_features = kin

            super(SimKFDecoderSup, self).load_decoder()



class SimKFDecoderShuffled(SimKFDecoder):
    '''
    Construct a KFDecoder based on encoder output in response to states simulated according to the state space model's process noise
    '''
    def load_decoder(self):
        '''
        Instantiate the neural encoder and "train" the decoder
        '''
        print("Creating simulation decoder..")
        encoder = self.encoder
        n_samples = 20000
        units = self.encoder.get_units()
        n_units = len(units)

        # draw samples from the W distribution
        ssm = self.ssm
        A, _, W = ssm.get_ssm_matrices()
        mean = np.zeros(A.shape[0])
        mean[-1] = 1
        state_samples = np.random.multivariate_normal(mean, 100*W, n_samples)

        spike_counts = np.zeros([n_units, n_samples])
        self.encoder.call_ds_rate = 1
        for k in range(n_samples):
            spike_counts[:,k] = np.array(self.encoder(state_samples[k])).ravel()

        inds = np.arange(spike_counts.shape[0])
        np.random.shuffle(inds)
        spike_counts = spike_counts[inds,:]

        kin = state_samples.T

        self.init_neural_features = spike_counts
        self.init_kin_features = kin

        self.decoder = train.train_KFDecoder_abstract(ssm, kin, spike_counts, units, 0.1)
        self.encoder.call_ds_rate = 6
        super(SimKFDecoderShuffled, self).load_decoder()

class SimKFDecoderRandom(SimKFDecoder):
    def load_decoder(self):
        '''
        Create a 'seed' decoder for the simulation which is simply randomly initialized
        '''
        from riglib.bmi import state_space_models
        ssm = state_space_models.StateSpaceEndptVel2D()
        self.decoder = train.rand_KFDecoder(ssm, self.encoder.get_units())
        super(SimKFDecoderRandom, self).load_decoder()


class SimPPFDecoderCursor(object):
    def load_decoder(self):
        # decoder_fname = '/storage/decoders/grom20150102_14_BPPF01021655.pkl'
        # decoder = pickle.load(open(decoder_fname))

        # beta = decoder.filt.C
        # self.init_beta = beta.copy()

        # inds = np.arange(beta.shape[0])
        # # np.random.shuffle(inds)
        # beta_dec = beta[inds, :]

        # dt = decoder.filt.dt 
        # self._init_neural_encoder()

        self.decoder = train._train_PPFDecoder_sim_known_beta(self.beta_full, self.encoder.units, dt=1./180)

class SimPPFDecoderCursorShuffled(object):
    def load_decoder(self):
        # decoder_fname = '/storage/decoders/grom20150102_14_BPPF01021655.pkl'
        # decoder = pickle.load(open(decoder_fname))

        # beta = decoder.filt.C
        # self.init_beta = beta.copy()

        inds = np.arange(self.beta_full.shape[0])
        np.random.shuffle(inds)
        self.shuffling_inds = inds
        # beta_dec = beta[inds, :]

        # dt = decoder.filt.dt 
        # self._init_neural_encoder()

        self.decoder = train._train_PPFDecoder_sim_known_beta(self.beta_full[inds], self.encoder.units, dt=1./180)



''' Tasks which don't include any visuals or bmi, such as laser-only or camera-only tasks'''

from riglib.experiment import LogExperiment, Sequence
from features.laser_features import DigitalWave
from riglib.experiment import traits
import itertools
import numpy as np

MAX_RECORD_EDGES = 32

class Conditions(Sequence):

    status = dict(
        wait = dict(start_trial="trial"),
        trial = dict(end_trial="wait", stoppable=False, end_state=True),
    )
    
    wait_time = traits.Float(5.0, desc="Inter-trial interval (s)")
    trial_time = traits.Float(1.0, desc="Trial duration (s)")
    sequence_generators = ['null_sequence']

    def init(self):
        self.trial_dtype = np.dtype([('trial', 'u4'), ('index', 'u4')])
        super().init()

    def _parse_next_trial(self):
        self.trial_index = self.next_trial

        # Send record of trial to sinks
        self.trial_record['trial'] = self.calc_trial_num()
        self.trial_record['index'] = self.trial_index
        self.sinks.send("trials", self.trial_record)

    def _test_start_trial(self, ts):
        return ts > self.wait_time and not self.pause
    
    def _test_end_trial(self, ts):
        return ts > self.trial_time

    def _start_trial(self):
        self.sync_event('TRIAL_START', self.trial_index)

    def _end_trial(self):
        self.sync_event('TRIAL_END')

    @classmethod
    def get_desc(cls, params, log_summary):
        duration = round(log_summary['runtime'] / 60, 1)
        return "{} trials in {} min".format(
            log_summary['n_trials'], duration)

    @staticmethod
    def gen_random_conditions(nreps, *args, replace=False):
        ''' Generate random sequence of all combinations of the given arguments'''
        unique = list(itertools.product(*args))
        conds = np.random.choice(nreps*len(unique), nreps*len(unique), replace=replace)
        seq = [[i % len(unique)] + list(unique[i % len(unique)]) for i in conds] # list of [index, arg1, arg2, ..., argn]
        return tuple(zip(*seq))

    @staticmethod
    def gen_conditions(nreps, *args, ascend=True):
        ''' Generate a sequential sequence of all combinations of the given arguments'''
        unique = list(itertools.product(*args))
        conds = np.tile(range(len(unique)), (nreps, 1)).T.flatten()
        if not ascend: # descending
            conds = np.flipud(conds)
        seq = [[i % len(unique)] + list(unique[i % len(unique)]) for i in conds] # list of [index, arg1, arg2, ..., argn]
        return tuple(zip(*seq))

    @staticmethod
    def null_sequence(ntrials=100):
        return [0 for _ in range(ntrials)]

class LaserConditions(Conditions):

    sequence_generators = ['single_laser_pulse', 'dual_laser_pulse', 'single_laser_square_wave', 'dual_laser_square_wave']
    exclude_parent_traits = ['trial_time']

    def __init__(self, *args, **kwargs):
        self.laser_threads = []
        super().__init__(*args, **kwargs)

    def init(self):
        self.trial_dtype = np.dtype([
            ('trial', 'u4'), 
            ('index', 'u4'),
            ('laser', 'S32'),
            ('power', 'f8'),
            ('edges', 'f8', MAX_RECORD_EDGES),
            ])
        super(Conditions, self).init()
    
    def run(self):
        if not (hasattr(self, 'lasers') and len(self.lasers) > 0):
            import io
            self.terminated_in_error = True
            self.termination_err = io.StringIO()
            self.termination_err.write("No laser feature enabled, cannot init LaserConditions")
            self.termination_err.seek(0)
            self.state = None
        super().run() 

    def _parse_next_trial(self):
        self.trial_index, self.laser_powers, self.laser_edges = self.next_trial
        if len(self.laser_powers) < len(self.lasers) or len(self.laser_edges) < len(self.lasers):
            raise AttributeError("Not enough laser sequences for the number of lasers enabled")
        
        # Set laser power
        for idx in range(len(self.lasers)):
            self.lasers[idx].set_power(self.laser_powers[idx])

        # Send record of trial to sinks
        self.trial_record['trial'] = self.calc_trial_num()
        self.trial_record['index'] = self.trial_index
        for idx in range(len(self.lasers)):
            self.trial_record['laser'] = self.lasers[idx].name
            self.trial_record['power'] = self.laser_powers[idx]
            record_laser_edges = self.laser_edges[idx][:MAX_RECORD_EDGES]
            self.trial_record['edges'] = np.pad(record_laser_edges, (0, MAX_RECORD_EDGES - len(record_laser_edges)), constant_values=np.nan)
            self.sinks.send("trials", self.trial_record)

    def _start_trial(self):
        super()._start_trial()
        for idx in range(len(self.lasers)):
            laser = self.lasers[idx]
            edges = self.laser_edges[idx]

            # Trigger digital wave
            wave = DigitalWave(laser, mask=1<<laser.port)
            wave.set_edges(edges, True)
            wave.start()
            self.laser_threads.append(wave)

    def _end_trial(self):
        super()._end_trial()
        # Turn laser off in between trials in case it ended on a rising edge
        for idx in range(len(self.lasers)):
            laser = self.lasers[idx]
            wave = DigitalWave(laser, mask=1>>laser.port)
            wave.set_edges([0], False)
            wave.start()
        
    def _test_end_trial(self, ts):
        return all([not t.is_alive() for t in self.laser_threads])

    @staticmethod
    def single_laser_pulse(nreps=100, duration=[0.005], power=[1], uniformsampling=True, ascending=False):
        '''
        Generates a sequence of laser pulse trains.

        Parameters
        ----------
        nreps : int
            The number of repetitions of each unique condition.
        duration: list of floats
            The duration of each pulse. Can be a list, randomly sampled
        power : list of floats
            Power for each pulse. Can be a list, randomly sampled

        Returns
        -------
        seq : (nreps*len(duration)*len(power) x 3) tuple of trial indices, laser powers, and edge sequences

        '''
        duration = make_list_of_float(duration)
        power = make_list_of_float(power)
        if uniformsampling:
            idx, dur_seq, pow_seq = Conditions.gen_random_conditions(nreps, duration, power)
        else:
            idx, dur_seq, pow_seq = Conditions.gen_conditions(nreps, duration, power, ascend=ascending)
        edge_seq = map(lambda dur: [0, dur], dur_seq)
        return list(zip(idx, [[p] for p in pow_seq], [[e] for e in edge_seq]))

    @staticmethod
    def dual_laser_pulse(nreps=100, dur_1=[0.005], power_1=[1], dur_2=[0.005], power_2=[1], uniformsampling=True, ascending=False):
        '''
        Generates a sequence of laser pulse trains that happen at the same time for two lasers.

        Parameters
        ----------
        nreps : int
            The number of repetitions of each unique condition.
        duration: list of floats
            The duration of each pulse. Can be a list, randomly sampled
        power : list of floats
            Power for each pulse. Can be a list, randomly sampled

        Returns
        -------
        seq : (nreps*len(duration)*len(power) x 3) tuple of trial indices, laser powers, and edge sequences

        '''
        dur_1 = make_list_of_float(dur_1)
        dur_2 = make_list_of_float(dur_2)
        power_1 = make_list_of_float(power_1)
        power_2 = make_list_of_float(power_2)
        if uniformsampling:
            idx, dur_1_seq, dur_2_seq, power_1_seq, power_2_seq = \
            Conditions.gen_random_conditions(nreps, dur_1, dur_2, power_1, power_2)
        else:
            idx, dur_1_seq, dur_2_seq, power_1_seq, power_2_seq = \
            Conditions.gen_conditions(nreps, dur_1, dur_2, power_1, power_2, ascend=ascending)
        edge_1_seq = map(lambda dur: [0, dur], dur_1_seq)
        edge_2_seq = map(lambda dur: [0, dur], dur_2_seq)
        return list(zip(idx, [[p1, p2] for p1, p2 in zip(power_1_seq, power_2_seq)], 
                             [[e1, e2] for e1, e2 in zip(edge_1_seq, edge_2_seq)]))


    @staticmethod
    def single_laser_square_wave(nreps=100, freq=[20], duration=[0.5], power=[1], uniformsampling=True, ascending=False):
        '''
        Generates a sequence of laser square waves.

        Parameters
        ----------
        nreps : int
            The number of repetitions of each unique condition.
        freq : list of floats
            The frequency for each square wave. Can be a list, randomly sampled
        duration: list of floats
            The duration of each square wave. Can be a list, randomly sampled
        power : list of floats
            Power for each square wave. Can be a list, randomly sampled

        Returns
        -------
        seq : (nreps*len(duration)*len(power)*len(freq) x 3) tuple of trial indices, laser powers, and edge sequences

        '''
        freq = make_list_of_float(freq)
        duration = make_list_of_float(duration)
        power = make_list_of_float(power)
        if uniformsampling:
            idx, freq_seq, dur_seq, pow_seq = Conditions.gen_random_conditions(nreps, freq, duration, power)
        else:
            idx, freq_seq, dur_seq, pow_seq = Conditions.gen_conditions(nreps, freq, duration, power, ascend=ascending)
        edge_seq = map(lambda freq, dur: DigitalWave.square_wave(freq, dur), freq_seq, dur_seq)
        return list(zip(idx, [[p] for p in pow_seq], [[e] for e in edge_seq]))

    @staticmethod
    def dual_laser_square_wave(nreps=100, freq_1=[20], freq_2=[20], dur_1=[0.5], dur_2=[0.5], duty_cycle_1=[0.5], duty_cycle_2=[0.5], 
        phase_delay_1=[0], phase_delay_2=[0], power_1=[1], power_2=[1], uniformsampling=True, ascending=False):
        '''
        Generates a sequence of laser square waves.

        Parameters
        ----------
        nreps : int
            The number of repetitions of each unique condition.
        freq : list of floats
            The frequency for each square wave. Can be a list, randomly sampled
        duration: list of floats
            The duration of each square wave. Can be a list, randomly sampled
        power : list of floats
            Power for each square wave. Can be a list, randomly sampled

        Returns
        -------
        seq : (nreps*len(duration)*len(power)*len(freq) x 3) tuple of trial indices, laser powers, and edge sequences

        '''
        freq_1 = make_list_of_float(freq_1)
        freq_2 = make_list_of_float(freq_2)
        dur_1 = make_list_of_float(dur_1)
        dur_2 = make_list_of_float(dur_2)
        duty_cycle_1 = make_list_of_float(duty_cycle_1)
        duty_cycle_2 = make_list_of_float(duty_cycle_2)
        phase_delay_1 = make_list_of_float(phase_delay_1)
        phase_delay_2 = make_list_of_float(phase_delay_2)
        power_1 = make_list_of_float(power_1)
        power_2 = make_list_of_float(power_2)
        if uniformsampling:
            idx, freq_1_seq, freq_2_seq, dur_1_seq, dur_2_seq, duty_cycle_1_seq, duty_cycle_2_seq, phase_delay_1_seq, phase_delay_2_seq, power_1_seq, power_2_seq = \
            Conditions.gen_random_conditions(nreps, freq_1, freq_2, dur_1, dur_2, duty_cycle_1, duty_cycle_2, phase_delay_1, phase_delay_2, power_1, power_2)
        else:
            idx, freq_1_seq, freq_2_seq, dur_1_seq, dur_2_seq, duty_cycle_1_seq, duty_cycle_2_seq, phase_delay_1_seq, phase_delay_2_seq, power_1_seq, power_2_seq = \
            Conditions.gen_conditions(nreps, freq_1, freq_2, dur_1, dur_2, duty_cycle_1, duty_cycle_2, phase_delay_1, phase_delay_2, power_1, power_2, ascend=ascending)
        edge_1_seq= map(lambda freq, dur, dc, delay: DigitalWave.square_wave(freq, dur, duty_cycle=dc, phase_delay=delay), freq_1_seq, dur_1_seq, duty_cycle_1_seq, phase_delay_1_seq)
        edge_2_seq = map(lambda freq, dur, dc, delay: DigitalWave.square_wave(freq, dur, duty_cycle=dc, phase_delay=delay), freq_2_seq, dur_2_seq, duty_cycle_2_seq, phase_delay_2_seq)
        return list(zip(idx, [[p1, p2] for p1, p2 in zip(power_1_seq, power_2_seq)], 
                             [[e1, e2] for e1, e2 in zip(edge_1_seq, edge_2_seq)]))


class SweptLaserConditions(LaserConditions):
    '''
    A task where conditions are swept over multiple lasers, one per trial.
    '''

    sequence_generators = ['swept_laser_pulse', 'swept_laser_square_wave']

    def _parse_next_trial(self):
        self.trial_index, self.laser_index, self.laser_power, self.laser_edges = self.next_trial

        # Set up laser
        self.lasers[self.laser_index].set_power(self.laser_power)
        self.lasers[self.laser_index].set_active()

        # Send record of trial to sinks
        self.trial_record['trial'] = self.calc_trial_num()
        self.trial_record['index'] = self.trial_index
        self.trial_record['laser'] = self.lasers[self.laser_index].name
        self.trial_record['power'] = self.laser_power
        record_laser_edges = self.laser_edges[:MAX_RECORD_EDGES]
        self.trial_record['edges'] = np.pad(record_laser_edges, (0, MAX_RECORD_EDGES - len(record_laser_edges)), constant_values=np.nan)
        self.sinks.send("trials", self.trial_record)

    def _start_trial(self):
        super(LaserConditions, self)._start_trial()
        laser = self.lasers[self.laser_index]
        edges = self.laser_edges

        # Trigger digital wave
        wave = DigitalWave(laser, mask=1<<laser.port)
        wave.set_edges(edges, True)
        wave.start()
        self.laser_threads.append(wave)

    def _end_trial(self):
        super(LaserConditions, self)._end_trial()

        # Turn laser off in between trials in case it ended on a rising edge
        laser = self.lasers[self.laser_index]
        wave = DigitalWave(laser, mask=1>>laser.port)
        wave.set_edges([0], False)
        wave.start()
        
    @staticmethod
    def swept_laser_pulse(nlasers=16, nreps=100, duration=[0.005], power=[1], uniformsampling=True, mixed_lasers=False):
        '''
        Generates a sequence of laser pulse trains.

        Parameters
        ----------
        nlasers : int
            The number of lasers to iterate over (can be more than the number of lasers enabled)
        nreps : int
            The number of repetitions of each unique condition.
        duration: list of floats
            The duration of each pulse. Can be a list, randomly sampled
        power : list of floats
            Power for each pulse. Can be a list, randomly sampled

        Returns
        -------
        seq : (nreps*len(duration)*len(power) x 3) tuple of trial indices, laser powers, and edge sequences

        '''
        duration = make_list_of_float(duration)
        power = make_list_of_float(power)
        lasers = range(nlasers)
        if mixed_lasers: 

            # Mix the lasers into the conditions
            if uniformsampling:
                idx, laser_seq, dur_seq, pow_seq = Conditions.gen_random_conditions(nreps, lasers, duration, power)
            else:
                idx, laser_seq, dur_seq, pow_seq = Conditions.gen_conditions(nreps, lasers, duration, power)
            edge_seq = map(lambda dur: [0, dur], dur_seq)
            return list(zip(idx, [ch for ch in laser_seq], [p for p in pow_seq], [e for e in edge_seq]))
        else:

            # Iterate over each laser
            if uniformsampling:
                idx, dur_seq, pow_seq = Conditions.gen_random_conditions(nreps, duration, power)
            else:
                idx, dur_seq, pow_seq = Conditions.gen_conditions(nreps, duration, power)
            edge_seq = list(map(lambda dur: [0, dur], dur_seq))
            full_list = []
            for ch in lasers:
                for idx in range(len(pow_seq)):
                    full_list.append((idx, ch, pow_seq[idx], edge_seq[idx]))
            return full_list
        
    @staticmethod
    def swept_laser_square_wave(nlasers=16, nreps=100, freq=[20], duration=[0.5], power=[1], uniformsampling=True, mixed_lasers=False):
        '''
        Generates a sequence of laser pulse trains.

        Parameters
        ----------
        nlasers : int
            The number of lasers to iterate over (can be more than the number of lasers enabled)
        nreps : int
            The number of repetitions of each unique condition.
        freq : list of floats
            The frequency for each square wave. Can be a list, randomly sampled
        duration: list of floats
            The duration of each pulse. Can be a list, randomly sampled
        power : list of floats
            Power for each pulse. Can be a list, randomly sampled

        Returns
        -------
        seq : (nreps*len(duration)*len(power) x 3) tuple of trial indices, laser powers, and edge sequences

        '''
        freq = make_list_of_float(freq)
        duration = make_list_of_float(duration)
        power = make_list_of_float(power)
        lasers = range(nlasers)
        if mixed_lasers: 

            # Mix the lasers into the conditions
            if uniformsampling:
                idx, laser_seq, freq_seq, dur_seq, pow_seq = Conditions.gen_random_conditions(nreps, lasers, freq, duration, power)
            else:
                idx, laser_seq, freq_seq, dur_seq, pow_seq = Conditions.gen_conditions(nreps, lasers, freq, duration, power)
            edge_seq = map(lambda freq, dur: DigitalWave.square_wave(freq, dur), freq_seq, dur_seq)
            return list(zip(idx, [ch for ch in laser_seq], [p for p in pow_seq], [e for e in edge_seq]))
        else:

            # Iterate over each laser
            if uniformsampling:
                idx, freq_seq, dur_seq, pow_seq = Conditions.gen_random_conditions(nreps, freq, duration, power)
            else:
                idx, freq_seq, dur_seq, pow_seq = Conditions.gen_conditions(nreps, freq, duration, power)
            edge_seq = list(map(lambda freq, dur: DigitalWave.square_wave(freq, dur), freq_seq, dur_seq))
            full_list = []
            for ch in lasers:
                for idx in range(len(pow_seq)):
                    full_list.append((idx, ch, pow_seq[idx], edge_seq[idx]))
            return full_list


####################
# Helper functions #
####################
def make_list_of_float(maybe_a_float):
    try:
        _ = iter(maybe_a_float)
    except TypeError:
        return [maybe_a_float]
    else:
        return maybe_a_float


'''
Peripheral interface device features
'''

import types
import numpy as np
import pygame
from riglib import gpio
from config.rig_defaults import force_sensor_address
from riglib.experiment import traits

###### CONSTANTS
sec_per_min = 60

class Joystick(traits.HasTraits):
    '''
    Grab x,y position data from a usb joystick using pygame
    '''
    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        self.joystick = JoystickInput(self.screen_cm)

class JoystickInput():
    '''
    Pretend to be a data source for a joystick. Scales the (-1,1) axis outputs to the screen dimensions
    '''

    def __init__(self, screen_cm):
        self.screen_cm = screen_cm

        # Enable polling the controller when the mouse isn't on the screen
        import os
        os.environ["SDL_JOYSTICK_ALLOW_BACKGROUND_EVENTS"] = "1"

        # Check if joystick is connected
        pygame.joystick.init()
        if pygame.joystick.get_count() == 0:
            raise Exception("No joystick connected!")
        self.joystick = pygame.joystick.Joystick(0)

    def get(self):
        pos = np.array([
            self.joystick.get_axis(0) * self.screen_cm[0],
            -self.joystick.get_axis(1) * self.screen_cm[1]
        ])
        return [pos]

class PhidgetsJoystick(object):
    '''
    Code to use an analog joystick with signals digitized by the phidgets board
    '''
    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' instantiates a DataSource with 2 channels for the two analog 
        inputs from the phidgets joystick. 
        '''
        from riglib import source, phidgets, sink
        sink_manager = sink.SinkManager.get_instance()
        #System = phidgets.make(2, 1)
        #self.joystick = source.DataSource(System)

        self.register_num_channels()
        super(Joystick, self).init()
        sink_manager.register(self.joystick)

    def register_num_channels(self):
        from riglib import source, phidgets, sink
        System = phidgets.make(2, 1)
        self.joystick = source.DataSource(System)

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the joystick source and stops it after the FSM has finished running
        '''
        self.joystick.start()
        try:
            super(Joystick, self).run()
        finally:
            self.joystick.stop()

    def join(self):
        '''
        See riglib.experiment.Experiment.join(). Re-join the joystick source process before cleaning up the experiment thread
        '''
        self.joystick.join()
        super(Joystick, self).join()

class ArduinoJoystick(Joystick):
    def init(self):
        '''
        Same as above, w/o Phidgets import
        '''
        from riglib import source, sink
        self.sinks = sink.sinks # sink_manager = sink.SinkManager.get_instance()
        self.register_num_channels()
        super(Joystick, self).init()
        self.sinks.register(self.joystick)

    def register_num_channels(self):
        from riglib import arduino_joystick, source, sink
        System = arduino_joystick.make(2, 1)
        self.joystick = source.DataSource(System)

class ArduinoIMU(object):
    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' instantiates a DataSource with 2 channels for the
        inputs from the IMU
        '''
        from riglib import sink
        self.register_num_channels()
        super(ArduinoIMU, self).init()

        sink_manager = sink.SinkManager.get_instance()
        sink_manager.register(self.arduino_imu)

    def register_num_channels(self):
        from riglib import source, arduino_imu
        System = arduino_imu.make(6, 1)
        self.arduino_imu = source.DataSource(System)

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the joystick source and stops it after the FSM has finished running
        '''
        self.arduino_imu.start()
        try:
            super(ArduinoIMU, self).run()
        finally:
            self.arduino_imu.stop()

    def join(self):
        '''
        See riglib.experiment.Experiment.join(). Re-join the joystick source process before cleaning up the experiment thread
        '''
        self.arduino_imu.join()
        super(ArduinoIMU, self).join()    

class Joystick_plus_TouchSensor(Joystick):
    '''
    code to use touch sensor (attached to joystick in exorig) plus joystick
    '''
    def register_num_channels(self):
        from riglib import source, phidgets, sink
        System = phidgets.make(3, 1)
        self.joystick = source.DataSource(System)

class DualJoystick(object):
    '''
    A two-joystick interface, similar to Joystick
    '''
    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' creates a 4-channel DataSource, two channels for each joystick
        -------
        '''
        from riglib import source, phidgets
        System = phidgets.make(4, 1)
        self.dualjoystick = source.DataSource(System)
        super(DualJoystick, self).init()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the dual_joystick source and stops it after the FSM has finished running
        '''
        self.dualjoystick.start()
        try:
            super(DualJoystick, self).run()
        finally:
            self.dualjoystick.stop()

    def join(self):
        '''
        See riglib.experiment.Experiment.join(). Re-join the joystick source process before cleaning up the experiment thread
        '''
        self.dualjoystick.join()
        super(DualJoystick, self).join()

class Button(object):
    '''
    Deprecated!

    Adds the ability to respond to the button, as well as to keyboard responses
    The "button" was a switch connected to a modified mouse so that the digital input went through 
    the mouse interface (hence the calls to pygame's mouse interface)
    '''
    def screen_init(self):
        super(Button, self).screen_init()
        import pygame
        pygame.event.set_grab(True)
        pygame.mouse.set_visible(False)

    def _get_event(self):
        import pygame
        btnmap = {1:1, 3:4}
        for btn in pygame.event.get(pygame.MOUSEBUTTONDOWN):
            if btn.button in btnmap:
                return btnmap[btn.button]

        return super(Button, self)._get_event()
    
    def _while_reward(self):
        super(Button, self)._while_reward()
        import pygame
        pygame.event.clear()
    
    def _while_penalty(self):
        #Clear out the button buffers
        super(Button, self)._while_penalty()
        import pygame
        pygame.event.clear()
    
    def _while_wait(self):
        super(Button, self)._while_wait()
        import pygame
        pygame.event.clear()

class EyeControl(object):
    '''
    this class implements a python cursor control task. This is just for testing eye related task. 
    '''

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        self.joystick = Eye(np.array(self.starting_pos[::2]))

class Eye(object):
    '''
    Pretend to be a data source. This is just for testing eye related task. 
    '''
    
    def __init__(self, start_pos):
        self.pos = [0., 0.]
        self.pos[0] = start_pos[0]
        self.pos[1] = start_pos[1]
        self.move_step = 1 # cm, before scaling
        self.calibration = np.array([2,2]) # TODO load calibration data

    def get(self):
        for event in pygame.event.get():
            if event.type == pygame.KEYUP:
                if event.type == pygame.K_q:
                    pygame.quit()
                    quit()
                if event.key == pygame.K_LEFT:
                    self.pos[0] -= self.move_step
                if event.key == pygame.K_RIGHT:
                    self.pos[0] += self.move_step
                if event.key == pygame.K_UP:
                    self.pos[1] += self.move_step
                if event.key == pygame.K_DOWN:
                    self.pos[1] -= self.move_step

        calibrated_pos = self.calibration*self.pos
        return [calibrated_pos]

class KeyboardControl(object):
    '''
    this class implements a python cursor control task for human
    '''

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        self.joystick = Keyboard(np.array(self.starting_pos[::2]))

class Keyboard():
    '''
    Pretend to be a data source
    '''

    def __init__(self, start_pos):
        self.pos = [0., 0.]
        self.pos[0] = start_pos[0]
        self.pos[1] = start_pos[1]
        self.move_step = 1 # cm, before scaling

    def get(self):
        for event in pygame.event.get():
            if event.type == pygame.KEYUP:
                if event.type == pygame.K_q:
                    pygame.quit()
                    quit()
                if event.key == pygame.K_LEFT:
                    self.pos[0] -= self.move_step
                if event.key == pygame.K_RIGHT:
                    self.pos[0] += self.move_step
                if event.key == pygame.K_UP:
                    self.pos[1] += self.move_step
                if event.key == pygame.K_DOWN:
                    self.pos[1] -= self.move_step
        return [self.pos]

class MouseControl(KeyboardControl):
    '''
    this class implements a python cursor control task for human
    '''

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        self.joystick = Mouse(self.window_size, self.screen_cm, np.array(self.starting_pos[::2]))

class Mouse():
    '''
    Pretend to be a data source
    '''

    def __init__(self, window_size, screen_cm, start_pos):
        self.window_size = window_size
        self.screen_cm = screen_cm
        self.pos = [0., 0.]
        self.pos[0] = start_pos[0]
        self.pos[1] = start_pos[1]

    def get(self):
        pos = pygame.mouse.get_pos()
        self.pos[0] = (pos[0] / self.window_size[0] - 0.5) * self.screen_cm[0]
        self.pos[1] = -(pos[1] / self.window_size[1] - 0.5) * self.screen_cm[1] # pygame counts (0,0) as the top left
        return [self.pos]
    
class ForceControl():
    '''
    Read the voltage from pin 0 on connected arduino at '/dev/forcesensor'
    '''
    def init(self, *args, **kwargs):
        if hasattr(super(), 'init'):
            super().init(*args, **kwargs)
        self.joystick = gpio.ArduinoGPIO(force_sensor_address, enable_analog=True)
        def get(self):
            return self.analog_read(0)
        self.joystick.get = types.MethodType(get, self.joystick)



# -*- coding: utf-8 -*-
"""
Collection of classes used for reading headers and data from Blackrock files
current version: 1.3.2 --- 08/12/2016

@author: Mitch Frankel - Blackrock Microsystems

Version History:
v1.0.0 - 07/05/2016 - initial release - requires brMiscFxns v1.0.0
v1.1.0 - 07/08/2016 - inclusion of NsxFile.savesubsetnsx() for saving subset of Nsx data to disk4
v1.1.1 - 07/09/2016 - update to NsxFile.savesubsetnsx() for option (not)overwriting subset files if already exist
                      bug fixes in NsxFile class as reported from beta user
v1.2.0 - 07/12/2016 - bug fixes in NsxFile.savesubsetnsx()
                      added version control and checking for brMiscFxns
                      requires brMiscFxns v1.1.0
v1.3.0 - 07/22/2016 - added 'samp_per_s' to NsxFile.getdata() output
                      added close() method to NsxFile and NevFile objects
                      NsxFile.getdata() now pre-allocates output['data'] as zeros - speed and safety
v1.3.1 - 08/02/2016 - bug fixes to NsxFile.getdata() for usability with Python 2.7 as reported from beta user
                      patch for use with multiple NSP sync (overwriting of initial null data from initial data packet)
                      __future__ import for use with Python 2.7 (division)
                      minor modifications to allow use of Python 2.6+
v1.3.2 - 08/12/2016 - bug fixes to NsXFile.getdata()
"""


  # for those using Python 2.6+
import numpy     as np
from collections import namedtuple
from datetime    import datetime
from math        import ceil
from os          import path as ospath
from struct      import calcsize, pack, unpack, unpack_from
from .brMiscFxns  import openfilecheck, brmiscfxns_ver

# Version control set/check
brpylib_ver        = "1.3.2"
brmiscfxns_ver_req = "1.2.0"
if brmiscfxns_ver.split('.') < brmiscfxns_ver_req.split('.'):
    raise Exception("brpylib requires brMiscFxns " + brmiscfxns_ver_req + " or higher, please use latest version")

# Patch for use with Python 2.6+
try: input = raw_input
except NameError: pass

# Define global variables to remove magic numbers
# <editor-fold desc="Globals">
WARNING_SLEEP_TIME      = 5
DATA_PAGING_SIZE        = 1024**3
DATA_FILE_SIZE_MIN      = 1024**2 * 10
STRING_TERMINUS         = '\x00'
UNDEFINED               = 0
ELEC_ID_DEF             = 'all'
START_TIME_DEF          = 0
DATA_TIME_DEF           = 'all'
DOWNSAMPLE_DEF          = 1
START_OFFSET_MIN        = 0
STOP_OFFSET_MIN         = 0

UV_PER_BIT_21             = 0.25
WAVEFORM_SAMPLES_21       = 48
NSX_BASIC_HEADER_BYTES_22 = 314
NSX_EXT_HEADER_BYTES_22   = 66
DATA_BYTE_SIZE            = 2
TIMESTAMP_NULL_21         = 0

NO_FILTER               = 0
BUTTER_FILTER           = 1
SERIAL_MODE             = 0

RB2D_MARKER             = 1
RB2D_BLOB               = 2
RB3D_MARKER             = 3
BOUNDARY_2D             = 4
MARKER_SIZE             = 5

DIGITAL_PACKET_ID       = 0
NEURAL_PACKET_ID_MIN    = 1
NEURAL_PACKET_ID_MAX    = 2048
COMMENT_PACKET_ID       = 65535
VIDEO_SYNC_PACKET_ID    = 65534
TRACKING_PACKET_ID      = 65533
BUTTON_PACKET_ID        = 65532
CONFIGURATION_PACKET_ID = 65531

PARALLEL_REASON         = 1
PERIODIC_REASON         = 64
SERIAL_REASON           = 129
LOWER_BYTE_MASK         = 255
FIRST_BIT_MASK          = 1
SECOND_BIT_MASK         = 2

CLASSIFIER_MIN          = 1
CLASSIFIER_MAX          = 16
CLASSIFIER_NOISE        = 255

CHARSET_ANSI            = 0
CHARSET_UTF             = 1
CHARSET_ROI             = 255

COMM_RGBA               = 0
COMM_TIME               = 1

BUTTON_PRESS            = 1
BUTTON_RESET            = 2

CHG_NORMAL              = 0
CHG_CRITICAL            = 1

ENTER_EVENT             = 1
EXIT_EVENT              = 2
# </editor-fold>

# Define a named tuple that has information about header/packet fields
FieldDef = namedtuple('FieldDef', ['name', 'formatStr', 'formatFnc'])


# <editor-fold desc="Header processing functions">
def processheaders(curr_file, packet_fields):
    """
    :param curr_file:      {file} the current BR datafile to be processed
    :param packet_fields : {named tuple} the specific binary fields for the given header
    :return:               a fully unpacked and formatted tuple set of header information

    Read a packet from a binary data file and return a list of fields
    The amount and format of data read will be specified by the
    packet_fields container
    """

    # This is a lot in one line.  First I pull out all the format strings from
    # the basic_header_fields named tuple, then concatenate them into a string
    # with '<' at the front (for little endian format)
    packet_format_str = '<' + ''.join([fmt for name, fmt, fun in packet_fields])

    # Calculate how many bytes to read based on the format strings of the header fields
    bytes_in_packet = calcsize(packet_format_str)
    packet_binary = curr_file.read(bytes_in_packet)

    # unpack the binary data from the header based on the format strings of each field.
    # This returns a list of data, but it's not always correctly formatted (eg, FileSpec
    # is read as ints 2 and 3 but I want it as '2.3'
    packet_unpacked = unpack(packet_format_str, packet_binary)

    # Create a iterator from the data list.  This allows a formatting function
    # to use more than one item from the list if needed, and the next formatting
    # function can pick up on the correct item in the list
    data_iter = iter(packet_unpacked)

    # create an empty dictionary from the name field of the packet_fields.
    # The loop below will fill in the values with formatted data by calling
    # each field's formatting function
    packet_formatted = dict.fromkeys([name for name, fmt, fun in packet_fields])
    for name, fmt, fun in packet_fields:
        packet_formatted[name] = fun(data_iter)

    return packet_formatted


def format_filespec(header_list):
    return str(next(header_list)) + '.' + str(next(header_list))  # eg 2.3


def format_timeorigin(header_list):
    year        = next(header_list)
    month       = next(header_list)
    _           = next(header_list)
    day         = next(header_list)
    hour        = next(header_list)
    minute      = next(header_list)
    second      = next(header_list)
    millisecond = next(header_list)
    return datetime(year, month, day, hour, minute, second, millisecond * 1000)


def format_stripstring(header_list):
    string = bytes.decode(next(header_list), 'latin-1')
    return string.split(STRING_TERMINUS, 1)[0]


def format_none(header_list):
    return next(header_list)


def format_freq(header_list):
    return str(float(next(header_list)) / 1000) + ' Hz'


def format_filter(header_list):
    filter_type = next(header_list)
    if filter_type == NO_FILTER:        return "none"
    elif filter_type == BUTTER_FILTER:  return "butterworth"


def format_charstring(header_list):
    return int(next(header_list))


def format_digconfig(header_list):
    config = next(header_list) & FIRST_BIT_MASK
    if config: return 'active'
    else:      return 'ignored'


def format_anaconfig(header_list):
    config = next(header_list)
    if config & FIRST_BIT_MASK:  return 'low_to_high'
    if config & SECOND_BIT_MASK: return 'high_to_low'
    else:                        return 'none'


def format_digmode(header_list):
    dig_mode = next(header_list)
    if dig_mode == SERIAL_MODE: return 'serial'
    else:                       return 'parallel'


def format_trackobjtype(header_list):
    trackobj_type = next(header_list)
    if   trackobj_type == UNDEFINED:    return 'undefined'
    elif trackobj_type == RB2D_MARKER:  return '2D RB markers'
    elif trackobj_type == RB2D_BLOB:    return '2D RB blob'
    elif trackobj_type == RB3D_MARKER:  return '3D RB markers'
    elif trackobj_type == BOUNDARY_2D:  return '2D boundary'
    elif trackobj_type == MARKER_SIZE:  return 'marker size'
    else:                               return 'error'


def getdigfactor(ext_headers, idx):
    max_analog  = ext_headers[idx]['MaxAnalogValue']
    min_analog  = ext_headers[idx]['MinAnalogValue']
    max_digital = ext_headers[idx]['MaxDigitalValue']
    min_digital = ext_headers[idx]['MinDigitalValue']
    return float(max_analog - min_analog) / float(max_digital - min_digital)
# </editor-fold>


# <editor-fold desc="Header dictionaries">
nev_header_dict = {
    'basic': [FieldDef('FileTypeID',            '8s',   format_stripstring),    # 8 bytes   - 8 char array
              FieldDef('FileSpec',              '2B',   format_filespec),       # 2 bytes   - 2 unsigned char
              FieldDef('AddFlags',              'H',    format_none),           # 2 bytes   - uint16
              FieldDef('BytesInHeader',         'I',    format_none),           # 4 bytes   - uint32
              FieldDef('BytesInDataPackets',    'I',    format_none),           # 4 bytes   - uint32
              FieldDef('TimeStampResolution',   'I',    format_none),           # 4 bytes   - uint32
              FieldDef('SampleTimeResolution',  'I',    format_none),           # 4 bytes   - uint32
              FieldDef('TimeOrigin',            '8H',   format_timeorigin),     # 16 bytes  - 8 x uint16
              FieldDef('CreatingApplication',   '32s',  format_stripstring),    # 32 bytes  - 32 char array
              FieldDef('Comment',               '256s', format_stripstring),    # 256 bytes - 256 char array
              FieldDef('NumExtendedHeaders',    'I',    format_none)],          # 4 bytes   - uint32

    'ARRAYNME': FieldDef('ArrayName',           '24s',  format_stripstring),    # 24 bytes  - 24 char array
    'ECOMMENT': FieldDef('ExtraComment',        '24s',  format_stripstring),    # 24 bytes  - 24 char array
    'CCOMMENT': FieldDef('ContComment',         '24s',  format_stripstring),    # 24 bytes  - 24 char array
    'MAPFILE':  FieldDef('MapFile',             '24s',  format_stripstring),    # 24 bytes  - 24 char array

    'NEUEVWAV': [FieldDef('ElectrodeID',        'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('PhysicalConnector',  'B',    format_charstring),     # 1 byte   - 1 unsigned char
                 FieldDef('ConnectorPin',       'B',    format_charstring),     # 1 byte   - 1 unsigned char
                 FieldDef('DigitizationFactor', 'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('EnergyThreshold',    'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('HighThreshold',      'h',    format_none),           # 2 bytes  - int16
                 FieldDef('LowThreshold',       'h',    format_none),           # 2 bytes  - int16
                 FieldDef('NumSortedUnits',     'B',    format_charstring),     # 1 byte   - 1 unsigned char
                 FieldDef('BytesPerWaveform',   'B',    format_charstring),     # 1 byte   - 1 unsigned char
                 FieldDef('SpikeWidthSamples',  'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('EmptyBytes',         '8s',   format_none)],          # 8 bytes  - empty

    'NEUEVLBL': [FieldDef('ElectrodeID',        'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('Label',              '16s',  format_stripstring),    # 16 bytes - 16 char array
                 FieldDef('EmptyBytes',         '6s',   format_none)],          # 6 bytes  - empty

    'NEUEVFLT': [FieldDef('ElectrodeID',        'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('HighFreqCorner',     'I',    format_freq),           # 4 bytes  - uint32
                 FieldDef('HighFreqOrder',      'I',    format_none),           # 4 bytes  - uint32
                 FieldDef('HighFreqType',       'H',    format_filter),         # 2 bytes  - uint16
                 FieldDef('LowFreqCorner',      'I',    format_freq),           # 4 bytes  - uint32
                 FieldDef('LowFreqOrder',       'I',    format_none),           # 4 bytes  - uint32
                 FieldDef('LowFreqType',        'H',    format_filter),         # 2 bytes  - uint16
                 FieldDef('EmptyBytes',         '2s',   format_none)],          # 2 bytes  - empty

    'DIGLABEL': [FieldDef('Label',              '16s',  format_stripstring),    # 16 bytes - 16 char array
                 FieldDef('Mode',               '?',    format_digmode),        # 1 byte   - boolean
                 FieldDef('EmptyBytes',         '7s',   format_none)],          # 7 bytes  - empty

    'NSASEXEV': [FieldDef('Frequency',          'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('DigitalInputConfig', 'B',    format_digconfig),      # 1 byte   - 1 unsigned char
                 FieldDef('AnalogCh1Config',    'B',    format_anaconfig),      # 1 byte   - 1 unsigned char
                 FieldDef('AnalogCh1DetectVal', 'h',    format_none),           # 2 bytes  - int16
                 FieldDef('AnalogCh2Config',    'B',    format_anaconfig),      # 1 byte   - 1 unsigned char
                 FieldDef('AnalogCh2DetectVal', 'h',    format_none),           # 2 bytes  - int16
                 FieldDef('AnalogCh3Config',    'B',    format_anaconfig),      # 1 byte   - 1 unsigned char
                 FieldDef('AnalogCh3DetectVal', 'h',    format_none),           # 2 bytes  - int16
                 FieldDef('AnalogCh4Config',    'B',    format_anaconfig),      # 1 byte   - 1 unsigned char
                 FieldDef('AnalogCh4DetectVal', 'h',    format_none),           # 2 bytes  - int16
                 FieldDef('AnalogCh5Config',    'B',    format_anaconfig),      # 1 byte   - 1 unsigned char
                 FieldDef('AnalogCh5DetectVal', 'h',    format_none),           # 2 bytes  - int16
                 FieldDef('EmptyBytes',         '6s',   format_none)],          # 2 bytes  - empty

    'VIDEOSYN': [FieldDef('VideoSourceID',      'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('VideoSource',        '16s',  format_stripstring),    # 16 bytes - 16 char array
                 FieldDef('FrameRate',          'f',    format_none),           # 4 bytes  - single float
                 FieldDef('EmptyBytes',         '2s',   format_none)],          # 2 bytes  - empty

    'TRACKOBJ': [FieldDef('TrackableType',      'H',    format_trackobjtype),   # 2 bytes  - uint16
                 FieldDef('TrackableID',        'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('PointCount',         'H',    format_none),           # 2 bytes  - uint16
                 FieldDef('VideoSource',        '16s',  format_stripstring),    # 16 bytes - 16 char array
                 FieldDef('EmptyBytes',         '2s',   format_none)]           # 2 bytes  - empty
}

nsx_header_dict = {
    'basic_21': [FieldDef('Label',              '16s', format_stripstring),   # 16 bytes  - 16 char array
                 FieldDef('Period',             'I',   format_none),          # 4 bytes   - uint32
                 FieldDef('ChannelCount',       'I',   format_none)],         # 4 bytes   - uint32

    'basic': [FieldDef('FileSpec',              '2B',   format_filespec),     # 2 bytes   - 2 unsigned char
              FieldDef('BytesInHeader',         'I',    format_none),         # 4 bytes   - uint32
              FieldDef('Label',                 '16s',  format_stripstring),  # 16 bytes  - 16 char array
              FieldDef('Comment',               '256s', format_stripstring),  # 256 bytes - 256 char array
              FieldDef('Period',                'I',    format_none),         # 4 bytes   - uint32
              FieldDef('TimeStampResolution',   'I',    format_none),         # 4 bytes   - uint32
              FieldDef('TimeOrigin',            '8H',   format_timeorigin),   # 16 bytes  - 8 uint16
              FieldDef('ChannelCount',          'I',    format_none)],        # 4 bytes   - uint32

    'extended': [FieldDef('Type',               '2s',   format_stripstring),  # 2 bytes   - 2 char array
                 FieldDef('ElectrodeID',        'H',    format_none),         # 2 bytes   - uint16
                 FieldDef('ElectrodeLabel',     '16s',  format_stripstring),  # 16 bytes  - 16 char array
                 FieldDef('PhysicalConnector',  'B',    format_none),         # 1 byte    - uint8
                 FieldDef('ConnectorPin',       'B',    format_none),         # 1 byte    - uint8
                 FieldDef('MinDigitalValue',    'h',    format_none),         # 2 bytes   - int16
                 FieldDef('MaxDigitalValue',    'h',    format_none),         # 2 bytes   - int16
                 FieldDef('MinAnalogValue',     'h',    format_none),         # 2 bytes   - int16
                 FieldDef('MaxAnalogValue',     'h',    format_none),         # 2 bytes   - int16
                 FieldDef('Units',              '16s',  format_stripstring),  # 16 bytes  - 16 char array
                 FieldDef('HighFreqCorner',     'I',    format_freq),         # 4 bytes   - uint32
                 FieldDef('HighFreqOrder',      'I',    format_none),         # 4 bytes   - uint32
                 FieldDef('HighFreqType',       'H',    format_filter),       # 2 bytes   - uint16
                 FieldDef('LowFreqCorner',      'I',    format_freq),         # 4 bytes   - uint32
                 FieldDef('LowFreqOrder',       'I',    format_none),         # 4 bytes   - uint32
                 FieldDef('LowFreqType',        'H',    format_filter)],      # 2 bytes   - uint16

    'data': [FieldDef('Header',                 'B',    format_none),         # 1 byte    - uint8
             FieldDef('Timestamp',              'I',    format_none),         # 4 bytes   - uint32
             FieldDef('NumDataPoints',          'I',    format_none)]         # 4 bytes   - uint32]
}
# </editor-fold>


# <editor-fold desc="Safety check functions">
def check_elecid(elec_ids):
    if type(elec_ids) is str and elec_ids != ELEC_ID_DEF:
        print("\n*** WARNING: Electrode IDs must be 'all', a single integer, or a list of integers.")
        print("      Setting elec_ids to 'all'")
        elec_ids = ELEC_ID_DEF
    if elec_ids != ELEC_ID_DEF and type(elec_ids) is not list:
        if type(elec_ids) == range: elec_ids = list(elec_ids)
        elif type(elec_ids) == int: elec_ids = [elec_ids]
    return elec_ids


def check_starttime(start_time_s):
    if not isinstance(start_time_s, (int, float)) or \
            (isinstance(start_time_s, (int, float)) and start_time_s < START_TIME_DEF):
        print("\n*** WARNING: Start time is not valid, setting start_time_s to 0")
        start_time_s = START_TIME_DEF
    return start_time_s


def check_datatime(data_time_s):
    if (type(data_time_s) is str and data_time_s != DATA_TIME_DEF) or \
            (isinstance(data_time_s, (int, float)) and data_time_s < 0):
        print("\n*** WARNING: Data time is not valid, setting data_time_s to 'all'")
        data_time_s = DATA_TIME_DEF
    return data_time_s


def check_downsample(downsample):
    if not isinstance(downsample, int) or downsample < DOWNSAMPLE_DEF:
        print("\n*** WARNING: Downsample must be an integer value greater than 0. "
              "      Setting downsample to 1 (no downsampling)")
        downsample = DOWNSAMPLE_DEF
    return downsample


def check_dataelecid(elec_ids, all_elec_ids):
    unique_elec_ids = set(elec_ids)
    all_elec_ids    = set(all_elec_ids)

    # if some electrodes asked for don't exist, reset list with those that do, or throw error and return
    if not unique_elec_ids.issubset(all_elec_ids):
        if not unique_elec_ids & all_elec_ids:
            print('\nNone of the elec_ids passed exist in the data, returning None')
            return None
        else:
            print(("\n*** WARNING: Channels " + str(sorted(list(unique_elec_ids - all_elec_ids))) +
                  " do not exist in the data"))
            unique_elec_ids = unique_elec_ids & all_elec_ids

    return sorted(list(unique_elec_ids))


def check_filesize(file_size):
    if file_size < DATA_FILE_SIZE_MIN:
        print('\n file_size must be larger than 10 Mb, setting file_size=10 Mb')
        return DATA_FILE_SIZE_MIN
    else:
        return int(file_size)
# </editor-fold>


class NevFile:
    """
    attributes and methods for all BR event data files.  Initialization opens the file and extracts the
    basic header information.
    """

    def __init__(self, datafile=''):
        self.datafile         = datafile
        self.basic_header     = {}
        self.extended_headers = []

        # Run openfilecheck and open the file passed or allow user to browse to one
        self.datafile = openfilecheck('rb', file_name=self.datafile, file_ext='.nev', file_type='Blackrock NEV Files')

        # extract basic header information
        self.basic_header = processheaders(self.datafile, nev_header_dict['basic'])

        # Extract extended headers
        for i in range(self.basic_header['NumExtendedHeaders']):
            self.extended_headers.append({})
            header_string = bytes.decode(unpack('<8s', self.datafile.read(8))[0], 'latin-1')
            self.extended_headers[i]['PacketID'] = header_string.split(STRING_TERMINUS, 1)[0]
            self.extended_headers[i].update(
                processheaders(self.datafile, nev_header_dict[self.extended_headers[i]['PacketID']]))

            # Must set this for file spec 2.1 and 2.2
            if header_string == 'NEUEVWAV' and float(self.basic_header['FileSpec']) < 2.3:
                self.extended_headers[i]['SpikeWidthSamples'] = WAVEFORM_SAMPLES_21

    def getdata(self, elec_ids='all'):
        """
        This function is used to return a set of data from the NSx datafile.

        :param elec_ids: [optional] {list} User selection of elec_ids to extract specific spike waveforms (e.g., [13])
        :return: output: {Dictionary} with one or more of the following dictionaries (all include TimeStamps)
                    dig_events:            Reason, Data, [for file spec 2.2 and below, AnalogData and AnalogDataUnits]
                    spike_events:          Units='nV', ChannelID, NEUEVWAV_HeaderIndices, Classification, Waveforms
                    comments:              CharSet, Flag, Data, Comment
                    video_sync_events:     VideoFileNum, VideoFrameNum, VideoElapsedTime_ms, VideoSourceID
                    tracking_events:       ParentID, NodeID, NodeCount, PointCount, TrackingPoints
                    button_trigger_events: TriggerType
                    configuration_events:  ConfigChangeType, ConfigChanged

        Note: For digital and neural data - TimeStamps, Classification, and Data can be lists of lists when more
        than one digital type or spike event exists for a channel
        """

        # Initialize output dictionary and reset position in file (if read before, may not be here anymore)
        output = dict()
        self.datafile.seek(self.basic_header['BytesInHeader'], 0)

        # Safety checks
        elec_ids = check_elecid(elec_ids)

        # Must go through each data packet and process separately until end of file
        while self.datafile.tell() != ospath.getsize(self.datafile.name):

            time_stamp = unpack('<I', self.datafile.read(4))[0]
            packet_id  = unpack('<H', self.datafile.read(2))[0]

            # skip unwanted neural data packets if only asking for certain channels
            if not (elec_ids == 'all' or ( (packet_id in elec_ids) and
                                           NEURAL_PACKET_ID_MIN <= packet_id <= NEURAL_PACKET_ID_MAX )):
                self.datafile.seek(self.basic_header['BytesInDataPackets'] - 6, 1)
                continue

            # For digital event data, read reason, skip one byte (reserved), read digital value,
            # and skip X bytes (reserved)
            if packet_id == DIGITAL_PACKET_ID:

                # See if the dictionary exists in output
                if 'dig_events' not in output:
                    output['dig_events'] = {'Reason': [], 'TimeStamps': [], 'Data': []}

                reason = unpack('B', self.datafile.read(1))[0]
                if   reason == PARALLEL_REASON: reason = 'parallel'
                elif reason == PERIODIC_REASON: reason = 'periodic'
                elif reason == SERIAL_REASON:   reason = 'serial'
                else:                           reason = 'unknown'
                self.datafile.seek(1, 1)

                # Check if this type of data already exists, if not, create an empty list, and then append data
                if reason in output['dig_events']['Reason']:
                    idx = output['dig_events']['Reason'].index(reason)
                else:
                    idx = -1
                    output['dig_events']['Reason'].append(reason)
                    output['dig_events']['TimeStamps'].append([])
                    output['dig_events']['Data'].append([])

                output['dig_events']['TimeStamps'][idx].append(time_stamp)
                output['dig_events']['Data'][idx].append(unpack('<H', self.datafile.read(2))[0])

                # For serial data, strip off upper byte
                if reason == 'serial':
                    output['dig_events']['Data'][idx][-1] &= LOWER_BYTE_MASK

                # For File Spec < 2.3, also capture analog Data, otherwise skip remaining packet bytes
                if float(self.basic_header['FileSpec']) < 2.3:
                    if 'AnalogDataUnits' not in output['dig_events']:
                        output['dig_events']['AnalogDataUnits'] = 'mv'

                    output['dig_events']['AnalogData'].append([])
                    for j in range(5):
                        output['dig_events']['AnalogData'][-1].append(unpack('<h', self.datafile.read(2))[0])
                else:
                    self.datafile.seek(self.basic_header['BytesInDataPackets'] - 10, 1)

            # For neural waveforms, read classifier, skip one byte (reserved), and read waveform data
            elif NEURAL_PACKET_ID_MIN <= packet_id <= NEURAL_PACKET_ID_MAX:

                # See if the dictionary exists in output, if not, create it
                if 'spike_events' not in output:
                    output['spike_events'] = {'Units': 'nV', 'ChannelID': [], 'TimeStamps': [],
                                              'NEUEVWAV_HeaderIndices': [], 'Classification': [], 'Waveforms': []}

                classifier = unpack('B', self.datafile.read(1))[0]
                if classifier == UNDEFINED:          classifier = 'none'
                elif CLASSIFIER_MIN <= classifier <= CLASSIFIER_MAX: classifier = classifier
                elif classifier == CLASSIFIER_NOISE: classifier = 'noise'
                else:                                classifier = 'error'
                self.datafile.seek(1, 1)

                # Check if data for this electrode exists and update parameters accordingly
                if packet_id in output['spike_events']['ChannelID']:
                    idx = output['spike_events']['ChannelID'].index(packet_id)
                else:
                    idx = -1
                    output['spike_events']['ChannelID'].append(packet_id)
                    output['spike_events']['TimeStamps'].append([])
                    output['spike_events']['Classification'].append([])

                    # Find neuevwav extended header for this electrode for use in calculating data info
                    output['spike_events']['NEUEVWAV_HeaderIndices'].append(
                        next(item for (item, d) in enumerate(self.extended_headers)
                             if d["ElectrodeID"] == packet_id and d["PacketID"] == 'NEUEVWAV'))

                output['spike_events']['TimeStamps'][idx].append(time_stamp)
                output['spike_events']['Classification'][idx].append(classifier)

                # Use extended header idx to get specific data information
                ext_header_idx = output['spike_events']['NEUEVWAV_HeaderIndices'][idx]
                samples    = self.extended_headers[ext_header_idx]['SpikeWidthSamples']
                dig_factor = self.extended_headers[ext_header_idx]['DigitizationFactor']
                num_bytes  = self.extended_headers[ext_header_idx]['BytesPerWaveform']
                if num_bytes <= 1:   data_type = np.int8
                elif num_bytes == 2: data_type = np.int16

                # Extract and scale the data
                if idx == -1:
                    output['spike_events']['Waveforms'].append(
                        [np.fromfile(file=self.datafile, dtype=data_type, count=samples).astype(np.int32) * dig_factor])
                else:
                    try:
                        output['spike_events']['Waveforms'][idx] = \
                            np.append(output['spike_events']['Waveforms'][idx],
                                [np.fromfile(file=self.datafile, dtype=data_type, count=samples).astype(np.int32) *
                                dig_factor], axis=0)
                    except:
                        output['spike_events']['Waveforms'][idx] = \
                            np.append(output['spike_events']['Waveforms'][idx],
                                [np.zeros((samples,)).astype(np.int32) * dig_factor], axis=0)
                        print('adding zero waveform')

            # For comment events
            elif packet_id == COMMENT_PACKET_ID:

                # See if the dictionary exists in output, if not, create it
                if 'comments' not in output:
                    output['comments'] = {'TimeStamps': [], 'CharSet': [], 'Flag': [], 'Data': [], 'Comment': []}

                output['comments']['TimeStamps'].append(time_stamp)

                char_set = unpack('B', self.datafile.read(1))[0]
                if char_set == CHARSET_ANSI:  output['comments']['CharSet'].append('ANSI')
                elif char_set == CHARSET_UTF: output['comments']['CharSet'].append('UTF-16')
                elif char_set == CHARSET_ROI: output['comments']['CharSet'].append('NeuroMotive ROI')
                else:                         output['comments']['CharSet'].append('error')

                comm_flag = unpack('B', self.datafile.read(1))[0]
                if comm_flag == COMM_RGBA:   output['comments']['Flag'].append('RGBA color code')
                elif comm_flag == COMM_TIME: output['comments']['Flag'].append('timestamp')
                else:                        output['comments']['Flag'].append('error')

                output['comments']['Data'].append(unpack('<I', self.datafile.read(4))[0])

                samples = self.basic_header['BytesInDataPackets'] - 12
                comm_string = bytes.decode(self.datafile.read(samples), 'latin-1')
                output['comments']['Comment'].append(comm_string.split(STRING_TERMINUS, 1)[0])

            # For video sync event
            elif packet_id == VIDEO_SYNC_PACKET_ID:

                # See if the dictionary exists in output, if not, create it
                if 'video_sync_events' not in output:
                    output['video_sync_events'] = {'TimeStamps': [], 'VideoFileNum': [], 'VideoFrameNum': [],
                                                   'VideoElapsedTime_ms': [], 'VideoSourceID': []}

                output['video_sync_events']['TimeStamps'].append(          time_stamp)
                output['video_sync_events']['VideoFileNum'].append(        unpack('<H', self.datafile.read(2))[0])
                output['video_sync_events']['VideoFrameNum'].append(       unpack('<I', self.datafile.read(4))[0])
                output['video_sync_events']['VideoElapsedTime_ms'].append( unpack('<I', self.datafile.read(4))[0])
                output['video_sync_events']['VideoSourceID'].append(       unpack('<I', self.datafile.read(4))[0])
                self.datafile.seek((self.basic_header['BytesInDataPackets'] - 20), 1)

            # For tracking event
            elif packet_id == TRACKING_PACKET_ID:

                # See if the dictionary exists in output, if not, create it
                if 'tracking_events' not in output:
                    output['tracking_events'] = {'TimeStamps': [], 'ParentID': [], 'NodeID': [], 'NodeCount': [],
                                                 'PointCount': [], 'TrackingPoints': []}

                output['tracking_events']['TimeStamps'].append( time_stamp)
                output['tracking_events']['ParentID'].append(   unpack('<H', self.datafile.read(2))[0])
                output['tracking_events']['NodeID'].append(     unpack('<H', self.datafile.read(2))[0])
                output['tracking_events']['NodeCount'].append(  unpack('<H', self.datafile.read(2))[0])
                output['tracking_events']['PointCount'].append( unpack('<H', self.datafile.read(2))[0])
                samples = (self.basic_header['BytesInDataPackets'] - 14) // 2
                output['tracking_events']['TrackingPoints'].append(
                    np.fromfile(file=self.datafile, dtype=np.uint16, count=samples))

            # For button trigger event
            elif packet_id == BUTTON_PACKET_ID:

                # See if the dictionary exists in output, if not, create it
                if 'button_trigger_events' not in output:
                    output['button_trigger_events'] = {'TimeStamps': [], 'TriggerType': []}

                output['button_trigger_events']['TimeStamps'].append(time_stamp)
                trigger_type = unpack('<H', self.datafile.read(2))[0]
                if trigger_type == UNDEFINED:      output['button_trigger_events']['TriggerType'].append('undefined')
                elif trigger_type == BUTTON_PRESS: output['button_trigger_events']['TriggerType'].append('button press')
                elif trigger_type == BUTTON_RESET: output['button_trigger_events']['TriggerType'].append('event reset')
                else:                              output['button_trigger_events']['TriggerType'].append('error')
                self.datafile.seek((self.basic_header['BytesInDataPackets'] - 8), 1)

            # For configuration log event
            elif packet_id == CONFIGURATION_PACKET_ID:

                # See if the dictionary exists in output, if not, create it
                if 'configuration_events' not in output:
                    output['configuration_events'] = {'TimeStamps': [], 'ConfigChangeType': [], 'ConfigChanged': []}

                output['configuration_events']['TimeStamps'].append(time_stamp)
                change_type = unpack('<H', self.datafile.read(2))[0]
                if change_type == CHG_NORMAL:     output['configuration_events']['ConfigChangeType'].append('normal')
                elif change_type == CHG_CRITICAL: output['configuration_events']['ConfigChangeType'].append('critical')
                else:                             output['configuration_events']['ConfigChangeType'].append('error')

                samples = self.basic_header['BytesInDataPackets'] - 8
                output['configuration_events']['ConfigChanged'].append(unpack(('<' + str(samples) + 's'),
                                                                              self.datafile.read(samples))[0])

            # Otherwise, packet unknown, skip to next packet
            else:  self.datafile.seek((self.basic_header['BytesInDataPackets'] - 6), 1)

        return output

    def processroicomments(self, comments):
        """
        used to process the comment data packets associated with NeuroMotive region of interest enter/exit events.
        requires that read_data() has already been run.
        :return: roi_events:   a dictionary of regions, enter timestamps, and exit timestamps for each region
        """

        roi_events = {'Regions': [], 'EnterTimeStamps': [], 'ExitTimeStamps': []}

        for i in range(len(comments['TimeStamps'])):
            if comments['CharSet'][i] == 'NeuroMotive ROI':

                temp_data = pack('<I', comments['Data'][i])
                roi   = unpack_from('<B', temp_data)[0]
                event = unpack_from('<B', temp_data, 1)[0]

                # Determine the label of the region source
                source_label = next(d['VideoSource'] for d in self.extended_headers if d["TrackableID"] == roi)

                # update the timestamps for events
                if source_label in roi_events['Regions']:
                    idx = roi_events['Regions'].index(source_label)
                else:
                    idx = -1
                    roi_events['Regions'].append(source_label)
                    roi_events['EnterTimeStamps'].append([])
                    roi_events['ExitTimeStamps'].append([])

                if   event == ENTER_EVENT: roi_events['EnterTimeStamps'][idx].append(comments['TimeStamp'][i])
                elif event == EXIT_EVENT:  roi_events['ExitTimeStamps'][idx].append(comments['TimeStamp'][i])

        return roi_events

    def close(self):
        name = self.datafile.name
        self.datafile.close()
        print(('\n' + name.split('/')[-1] + ' closed'))


class NsxFile:
    """
    attributes and methods for all BR continuous data files.  Initialization opens the file and extracts the
    basic header information.
    """

    def __init__(self, datafile=''):

        self.datafile         = datafile
        self.basic_header     = {}
        self.extended_headers = []

        # Run openfilecheck and open the file passed or allow user to browse to one
        self.datafile = openfilecheck('rb', file_name=self.datafile, file_ext='.ns*', file_type='Blackrock NSx Files')

        # Determine File ID to determine if File Spec 2.1
        self.basic_header['FileTypeID'] = bytes.decode(self.datafile.read(8), 'latin-1')

        # Extract basic and extended header information based on File Spec
        if self.basic_header['FileTypeID'] == 'NEURALSG':
            self.basic_header.update(processheaders(self.datafile, nsx_header_dict['basic_21']))
            self.basic_header['FileSpec']            = '2.1'
            self.basic_header['TimeStampResolution'] = 30000
            self.basic_header['BytesInHeader']       = 32 + 4 * self.basic_header['ChannelCount']
            shape = (1, self.basic_header['ChannelCount'])
            self.basic_header['ChannelID'] = \
                list(np.fromfile(file=self.datafile, dtype=np.uint32,
                                 count=self.basic_header['ChannelCount']).reshape(shape)[0])
        else:
            self.basic_header.update(processheaders(self.datafile, nsx_header_dict['basic']))
            for i in range(self.basic_header['ChannelCount']):
                self.extended_headers.append(processheaders(self.datafile, nsx_header_dict['extended']))

    def getdata(self, elec_ids='all', start_time_s=0, data_time_s='all', downsample=1):
        """
        This function is used to return a set of data from the NSx datafile.

        :param elec_ids:      [optional] {list}  List of elec_ids to extract (e.g., [13])
        :param start_time_s:  [optional] {float} Starting time for data extraction (e.g., 1.0)
        :param data_time_s:   [optional] {float} Length of time of data to return (e.g., 30.0)
        :param downsample:    [optional] {int}   Downsampling factor (e.g., 2)
        :return: output:      {Dictionary} of:  data_headers: {list}        dictionaries of all data headers
                                                elec_ids:     {list}        elec_ids that were extracted (sorted)
                                                start_time_s: {float}       starting time for data extraction
                                                data_time_s:  {float}       length of time of data returned
                                                downsample:   {int}         data downsampling factor
                                                samp_per_s:   {float}       output data samples per second
                                                data:         {numpy array} continuous data in a 2D numpy array

        Parameters: elec_ids, start_time_s, data_time_s, and downsample are not mandatory.  Defaults will assume all
        electrodes and all data points starting at time(0) are to be read. Data is returned as a numpy 2d array
        with each row being the data set for each electrode (e.g. output['data'][0] for output['elec_ids'][0]).
        """

        # Safety checks
        start_time_s = check_starttime(start_time_s)
        data_time_s  = check_datatime(data_time_s)
        downsample   = check_downsample(downsample)
        elec_ids     = check_elecid(elec_ids)

        # initialize parameters
        output                          = dict()
        output['elec_ids']              = elec_ids
        output['start_time_s']          = float(start_time_s)
        output['data_time_s']           = data_time_s
        output['downsample']            = downsample
        output['data']                  = []
        output['data_headers']          = []
        output['ExtendedHeaderIndices'] = []

        datafile_samp_per_sec = self.basic_header['TimeStampResolution'] / self.basic_header['Period']
        data_pt_size          = self.basic_header['ChannelCount'] * DATA_BYTE_SIZE
        elec_id_indices       = []
        front_end_idxs        = []
        analog_input_idxs     = []
        front_end_idx_cont    = True
        analog_input_idx_cont = True
        hit_start             = False
        hit_stop              = False
        d_ptr                 = 0

        # Move file position to start of datafile (if read before, may not be here anymore)
        self.datafile.seek(self.basic_header['BytesInHeader'], 0)

        # Based on FileSpec set other parameters
        if self.basic_header['FileSpec'] == '2.1':
            output['elec_ids'] = self.basic_header['ChannelID']
            output['data_headers'].append({})
            output['data_headers'][0]['Timestamp']     = TIMESTAMP_NULL_21
            output['data_headers'][0]['NumDataPoints'] = (ospath.getsize(self.datafile.name) - self.datafile.tell()) \
                                                         // (DATA_BYTE_SIZE * self.basic_header['ChannelCount'])
        else:
            output['elec_ids'] = [d['ElectrodeID'] for d in self.extended_headers]

        # Determine start and stop index for data
        if start_time_s == START_TIME_DEF: start_idx = START_OFFSET_MIN
        else:                              start_idx = int(round(start_time_s * datafile_samp_per_sec))
        if data_time_s == DATA_TIME_DEF:   stop_idx  = STOP_OFFSET_MIN
        else:                              stop_idx  = int(round((start_time_s + data_time_s) * datafile_samp_per_sec))

        # If a subset of electrodes is requested, error check, determine elec indices, and reduce headers
        if elec_ids != ELEC_ID_DEF:
            elec_ids = check_dataelecid(elec_ids, output['elec_ids'])
            if not elec_ids: return output
            else:
                elec_id_indices    = [output['elec_ids'].index(e) for e in elec_ids]
                output['elec_ids'] = elec_ids
        num_elecs = len(output['elec_ids'])

        # Determine extended header indices and idx for Front End vs. Analog Input channels
        if self.basic_header['FileSpec'] != '2.1':
            for i in range(num_elecs):
                idx = next(item for (item, d) in enumerate(self.extended_headers)
                           if d["ElectrodeID"] == output['elec_ids'][i])
                output['ExtendedHeaderIndices'].append(idx)

                if self.extended_headers[idx]['PhysicalConnector'] < 5: front_end_idxs.append(i)
                else:                                                   analog_input_idxs.append(i)

            # Determine if front_end_idxs and analog_idxs are contiguous (default = False)
            if any(np.diff(np.array(front_end_idxs)) != 1):     front_end_idx_cont    = False
            if any(np.diff(np.array(analog_input_idxs)) != 1):  analog_input_idx_cont = False

        # Pre-allocate output data based on data packet info (timestamp + num pts) and/or data_time_s
        # 1) Determine number of samples in all data packets to set possible number of output pts
        # 1a) For file spec > 2.1, get to last data packet quickly to determine total possible output length
        # 2) If possible output length is bigger than requested, set output based on requested
        if self.basic_header['FileSpec'] == '2.1':
            timestamp    = TIMESTAMP_NULL_21
            num_data_pts = output['data_headers'][0]['NumDataPoints']
        else:
            while self.datafile.tell() != ospath.getsize(self.datafile.name):
                self.datafile.seek(1, 1)  # skip header byte value
                timestamp    = unpack('<I', self.datafile.read(4))[0]
                num_data_pts = unpack('<I', self.datafile.read(4))[0]
                self.datafile.seek(num_data_pts * self.basic_header['ChannelCount'] * DATA_BYTE_SIZE, 1)

        stop_idx_output = ceil(timestamp / self.basic_header['Period']) + num_data_pts
        if data_time_s != DATA_TIME_DEF and stop_idx < stop_idx_output:  stop_idx_output = stop_idx
        total_samps = int(ceil((stop_idx_output - start_idx) / downsample))

        if (total_samps * self.basic_header['ChannelCount'] * DATA_BYTE_SIZE) > DATA_PAGING_SIZE:
            print("\nOutput data requested is larger than 1 GB, attempting to preallocate output now")

        # If data output is bigger than available, let user know this is too big and they must request at least one of:
        # subset of electrodes, subset of data, or use savensxsubset to smaller file sizes, otherwise, pre-allocate data
        try:   output['data'] = np.zeros((total_samps, num_elecs), dtype=np.float32)
        except MemoryError as err:
            err.args += (" Output data size requested is larger than available memory. Use the parameters\n"
                         "              for getdata(), e.g., 'elec_ids', to request a subset of the data or use\n"
                         "              NsxFile.savesubsetnsx() to create subsets of the main nsx file\n", )
            raise

        # Reset file position to start of data header #1, loop through all data packets, process header, and add data
        self.datafile.seek(self.basic_header['BytesInHeader'], 0)
        while not hit_stop:

            # Read header, check to make sure the header is valid (ie Header field != 0).  There is currently a
            # bug with the NSP where pausing creates a 0 sample packet before the next real data packet, these need to
            # be skipped, including any tiny packets that have less samples than downsample
            if self.basic_header['FileSpec'] != '2.1':
                output['data_headers'].append(processheaders(self.datafile, nsx_header_dict['data']))
                if output['data_headers'][-1]['Header'] == 0:  print('Invalid Header.  File may be corrupt')
                if output['data_headers'][-1]['NumDataPoints'] < downsample:
                    self.datafile.seek(self.basic_header['ChannelCount'] * output['data_headers'][-1]['NumDataPoints']
                                       * DATA_BYTE_SIZE, 1)
                    continue

            # Determine sample value for current packet timestamp
            timestamp_sample = int(round(output['data_headers'][-1]['Timestamp'] / self.basic_header['Period']))

            # For now, we need a patch for file sync which syncs 2 NSP clocks, starting a new data packet which
            # may be backwards in time wrt the end of data packet 1.  Thus, when this happens, we need to treat
            # data packet 2 as if it was 1, and start this process over.
            if timestamp_sample < d_ptr:
                d_ptr = 0
                hit_start = False
                output['data_headers'] = []
                self.datafile.seek(-9, 1)
                continue

            # Check to see if stop index is before the first data packet
            if len(output['data_headers']) == 1 and (STOP_OFFSET_MIN < stop_idx < timestamp_sample):
                print(("\nData requested is before any data was saved, which starts at t = {0:.6f} s".format(
                    output['data_headers'][0]['Timestamp'] / self.basic_header['TimeStampResolution'])))
                return

            # For the first data packet to be read
            if not hit_start:

                # Check for starting point of data request
                start_offset = start_idx - timestamp_sample

                # If start_offset is outside of this packet, skip the current packet
                # if we've reached the end of file, break, otherwise continue to next packet
                if start_offset > output['data_headers'][-1]['NumDataPoints']:
                    self.datafile.seek(output['data_headers'][-1]['NumDataPoints'] * data_pt_size, 1)
                    if self.datafile.tell() == ospath.getsize(self.datafile.name): break
                    else: continue

                else:
                    # If the start_offset is before the current packet, check to ensure that stop_index
                    # is not also in the paused area, then create padded data for during pause time
                    if start_offset < 0:
                        if STOP_OFFSET_MIN < stop_idx < timestamp_sample:
                            print("\nBecause of pausing, data section requested is during pause period")
                            return
                        else:
                            print(("\nFirst data packet requested begins at t = {0:.6f} s, "
                                  "initial section padded with zeros".format(
                                   output['data_headers'][-1]['Timestamp'] / self.basic_header['TimeStampResolution'])))
                            start_offset = START_OFFSET_MIN
                            d_ptr        = (timestamp_sample - start_idx) // downsample
                    hit_start = True

            # for all other packets
            else:
                # check to see if padded data is needed, including hitting the stop index
                if STOP_OFFSET_MIN < stop_idx < timestamp_sample:
                    print("\nSection padded with zeros due to file pausing")
                    hit_stop = True;  break

                elif (timestamp_sample - start_idx) > d_ptr:
                    print("\nSection padded with zeros due to file pausing")
                    start_offset = START_OFFSET_MIN
                    d_ptr        = (timestamp_sample - start_idx) // downsample

            # Set number of samples to be read based on if start/stop sample is during data packet
            if STOP_OFFSET_MIN < stop_idx <= (timestamp_sample + output['data_headers'][-1]['NumDataPoints']):
                total_pts = stop_idx - timestamp_sample - start_offset
                hit_stop = True
            else:
                total_pts = output['data_headers'][-1]['NumDataPoints'] - start_offset

            # Need current file position because memory map will reset file position
            curr_file_pos = self.datafile.tell()

            # Determine starting position to read from memory map
            file_offset = int(curr_file_pos + start_offset * data_pt_size)

            # Extract data no more than 1 GB at a time (or based on DATA_PAGING_SIZE)
            # Determine shape of data to map based on file sizing and position, then map it
            downsample_data_size = data_pt_size * downsample
            max_length           = (DATA_PAGING_SIZE // downsample_data_size) * downsample_data_size
            num_loops            = int(ceil(total_pts * data_pt_size / max_length))

            for loop in range(num_loops):
                if loop == 0:
                    if num_loops == 1:  num_pts = total_pts
                    else:               num_pts = max_length // data_pt_size

                else:
                    file_offset += max_length
                    if loop == (num_loops - 1):  num_pts = ((total_pts * data_pt_size) % max_length) // data_pt_size
                    else:                        num_pts = max_length // data_pt_size
                        
                if num_loops != 1:  print(('Data extraction requires paging: {0} of {1}'.format(loop + 1, num_loops)))

                num_pts = int(num_pts)
                shape   = (num_pts, self.basic_header['ChannelCount'])
                mm      = np.memmap(self.datafile, dtype=np.int16, mode='r', offset=file_offset, shape=shape)

                # append data based on downsample slice and elec_ids indexing, then clear memory map
                if downsample != 1: mm = mm[::downsample]
                if elec_id_indices:
                    output['data'][d_ptr:d_ptr + mm.shape[0]] = np.array(mm[:, elec_id_indices]).astype(np.float32)
                else:
                    output['data'][d_ptr:d_ptr + mm.shape[0]] = np.array(mm).astype(np.float32)
                d_ptr += mm.shape[0]
                del mm

            # Reset current file position for file position checking and possibly next header
            curr_file_pos += self.basic_header['ChannelCount'] * output['data_headers'][-1]['NumDataPoints'] \
                             * DATA_BYTE_SIZE
            self.datafile.seek(curr_file_pos, 0)
            if curr_file_pos == ospath.getsize(self.datafile.name):  hit_stop = True

        # Safety checks for start and stop times
        if not hit_stop and start_idx > START_OFFSET_MIN:
            raise Exception('Error: End of file found before start_time_s')
        elif not hit_stop and stop_idx:
            print("\n*** WARNING: End of file found before stop_time_s, returning all data in file")

        # Transpose the data so that it has entries based on each electrode, not each sample time
        output['data'] = output['data'].transpose()

        # All data must be scaled based on scaling factors from extended header
        if self.basic_header['FileSpec'] == '2.1':  output['data'] *= UV_PER_BIT_21
        else:
            if front_end_idxs:
                if front_end_idx_cont:
                    output['data'][front_end_idxs[0]:front_end_idxs[-1] + 1] *= \
                        getdigfactor(self.extended_headers, output['ExtendedHeaderIndices'][front_end_idxs[0]])
                else:
                    for i in front_end_idxs:
                        output['data'][i] *= getdigfactor(self.extended_headers, output['ExtendedHeaderIndices'][i])

            if analog_input_idxs:
                if analog_input_idx_cont:
                    output['data'][analog_input_idxs[0]:analog_input_idxs[-1] + 1] *= \
                        getdigfactor(self.extended_headers, output['ExtendedHeaderIndices'][analog_input_idxs[0]])
                else:
                    for i in analog_input_idxs:
                        output['data'][i] *= getdigfactor(self.extended_headers, output['ExtendedHeaderIndices'][i])

        # Update parameters based on data extracted
        output['samp_per_s']  = float(datafile_samp_per_sec / downsample)
        output['data_time_s'] = len(output['data'][0]) / output['samp_per_s']

        return output

    def savesubsetnsx(self, elec_ids='all', file_size=None, file_time_s=None, file_suffix=''):
        """
        This function is used to save a subset of data based on electrode IDs, file sizing, or file data time.  If
        both file_time_s and file_size are passed, it will default to file_time_s and determine sizing accordingly.

        :param elec_ids:    [optional] {list}  List of elec_ids to extract (e.g., [13])
        :param file_size:   [optional] {int}   Byte size of each subset file to save (e.g., 1024**3 = 1 Gb). If nothing
                                                   is passed, file_size will be all data points.
        :param file_time_s: [optional] {float} Time length of data for each subset file, in seconds (e.g. 60.0).  If
                                                   nothing is passed, file_size will be used as default.
        :param file_suffix: [optional] {str}   Suffix to append to NSx datafile name for subset files.  If nothing is
                                                   passed, default will be "_subset".
        :return: None - None of the electrodes requested exist in the data
                 SUCCESS - All file subsets extracted and saved
        """

        # Initializations
        elec_id_indices      = []
        file_num             = 1
        pausing              = False
        datafile_datapt_size = self.basic_header['ChannelCount'] * DATA_BYTE_SIZE
        self.datafile.seek(0, 0)

        # Run electrode id checks and set num_elecs
        elec_ids = check_elecid(elec_ids)
        if self.basic_header['FileSpec'] == '2.1': all_elec_ids = self.basic_header['ChannelID']
        else:                                      all_elec_ids = [x['ElectrodeID'] for x in self.extended_headers]

        if elec_ids == ELEC_ID_DEF:
            elec_ids = all_elec_ids
        else:
            elec_ids = check_dataelecid(elec_ids, all_elec_ids)
            if not elec_ids: return None
            else:            elec_id_indices = [all_elec_ids.index(x) for x in elec_ids]

        num_elecs = len(elec_ids)

        # If file_size or file_time_s passed, check it and set file_sizing accordingly
        if file_time_s:
            if file_time_s and file_size:
                print("\nWARNING: Only one of file_size or file_time_s can be passed, defaulting to file_time_s.")
            file_size = int(num_elecs * DATA_BYTE_SIZE * file_time_s *
                            self.basic_header['TimeStampResolution'] / self.basic_header['Period'])
            if self.basic_header['FileSpec'] == '2.1':
                file_size += 32 + 4 * num_elecs
            else:
                file_size += NSX_BASIC_HEADER_BYTES_22 + NSX_EXT_HEADER_BYTES_22 * num_elecs + 5
            print(("\nBased on timing request, file size will be {0:d} Mb".format(int(file_size / 1024**2))))
        elif file_size:
            file_size = check_filesize(file_size)

        # Create and open subset file as writable binary, if it already exists ask user for overwrite permission
        file_name, file_ext = ospath.splitext(self.datafile.name)
        if file_suffix:  file_name += '_' + file_suffix
        else:            file_name += '_subset'

        if ospath.isfile(file_name + "_000" + file_ext):
            if 'y' != eval(input("\nFile '" + file_name.split('/')[-1] + "_xxx" + file_ext +
                                    "' already exists, overwrite [y/n]: ")):
                print("\nExiting, no overwrite, returning None"); return None
            else:
                print("\n*** Overwriting existing subset files ***")

        subset_file = open(file_name + "_000" + file_ext, 'wb')
        print(("\nWriting subset file: " + ospath.split(subset_file.name)[1]))

        # For file spec 2.1:
        #   1) copy the first 28 bytes from the datafile (these are unchanged)
        #   2) write subset channel count and channel ID to file
        #   3) skip ahead in datafile the number of bytes in datafile ChannelCount(4) plus ChannelID (4*ChannelCount)
        if self.basic_header['FileSpec'] == '2.1':
            subset_file.write(self.datafile.read(28))
            subset_file.write(np.array(num_elecs).astype(np.uint32).tobytes())
            subset_file.write(np.array(elec_ids).astype(np.uint32).tobytes())
            self.datafile.seek(4 + 4 * self.basic_header['ChannelCount'], 1)

        # For file spec 2.2 and above
        #    1) copy the first 10 bytes from the datafile (unchanged)
        #    2) write subset bytes-in-headers and skip 4 bytes in datafile, noting position of this for update later
        #    3) copy the next 296 bytes from datafile (unchanged)
        #    4) write subset channel-count value and skip 4 bytes in datafile
        #    5) append extended headers based on the channel ID.  Must read the first 4 bytes, determine if correct
        #          Channel ID, repack first 4 bytes, write to disk, then copy remaining 62 (66-4) bytes
        else:
            subset_file.write(self.datafile.read(10))
            bytes_in_headers = NSX_BASIC_HEADER_BYTES_22 + NSX_EXT_HEADER_BYTES_22 * num_elecs
            num_pts_header_pos = bytes_in_headers + 5
            subset_file.write(np.array(bytes_in_headers).astype(np.uint32).tobytes())
            self.datafile.seek(4, 1)
            subset_file.write(self.datafile.read(296))
            subset_file.write(np.array(num_elecs).astype(np.uint32).tobytes())
            self.datafile.seek(4, 1)

            for i in range(len(self.extended_headers)):
                h_type  = self.datafile.read(2)
                chan_id = self.datafile.read(2)
                if unpack('<H', chan_id)[0] in elec_ids:
                    subset_file.write(h_type)
                    subset_file.write(chan_id)
                    subset_file.write(self.datafile.read(62))
                else:
                    self.datafile.seek(62, 1)

        # For all file types, loop through all data packets, extracting data based on page sizing
        while self.datafile.tell() != ospath.getsize(self.datafile.name):

            # pull and set data packet header info
            if self.basic_header['FileSpec'] == '2.1':
                packet_pts = (ospath.getsize(self.datafile.name) - self.datafile.tell()) \
                             / (DATA_BYTE_SIZE * self.basic_header['ChannelCount'])
            else:
                header_binary     = self.datafile.read(1)
                timestamp_binary  = self.datafile.read(4)
                packet_pts_binary = self.datafile.read(4)
                packet_pts        = unpack('<I', packet_pts_binary)[0]
                if packet_pts == 0: continue

                subset_file.write(header_binary)
                subset_file.write(timestamp_binary)
                subset_file.write(packet_pts_binary)

            # get current file position and set loop parameters
            datafile_pos        = self.datafile.tell()
            file_offset         = datafile_pos
            mm_length           = (DATA_PAGING_SIZE // datafile_datapt_size) * datafile_datapt_size
            num_loops           = int(ceil(packet_pts * datafile_datapt_size / mm_length))
            packet_read_pts     = 0
            subset_file_pkt_pts = 0

            # Determine shape of data to map based on file sizing and position, map it, then append to file
            for loop in range(num_loops):
                if loop == 0:
                    if num_loops == 1:  num_pts = packet_pts
                    else:               num_pts = mm_length // datafile_datapt_size

                else:
                    file_offset += mm_length
                    if loop == (num_loops - 1):
                        num_pts = ((packet_pts * datafile_datapt_size) % mm_length) // datafile_datapt_size
                    else:
                        num_pts = mm_length // datafile_datapt_size

                shape = (int(num_pts), self.basic_header['ChannelCount'])
                mm = np.memmap(self.datafile, dtype=np.int16, mode='r', offset=file_offset, shape=shape)
                if elec_id_indices: mm = mm[:, elec_id_indices]
                start_idx = 0

                # Determine if we need to start an additional file
                if file_size and (file_size - subset_file.tell()) < DATA_PAGING_SIZE:

                    # number of points we can possibly write to current subset file
                    pts_can_add = int((file_size - subset_file.tell()) // (num_elecs * DATA_BYTE_SIZE)) + 1
                    stop_idx    = start_idx + pts_can_add

                    # If the pts remaining are less than exist in the data, we'll need an additional subset file
                    while pts_can_add < num_pts:

                        # Write pts to disk, set old file name, update pts in packet, and close last subset file
                        if elec_id_indices:  subset_file.write(np.array(mm[start_idx:stop_idx]).tobytes())
                        else:                subset_file.write(mm[start_idx:stop_idx])
                        prior_file_name    = subset_file.name
                        prior_file_pkt_pts = subset_file_pkt_pts + pts_can_add
                        subset_file.close()

                        # We need to copy header information from last subset file and adjust some headers.
                        # For file spec 2.1, this is just the basic header.
                        # For file spec 2.2 and above:
                        #    1) copy basic and extended headers
                        #    2) create data packet header with new timestamp and num data points (dummy numpts value)
                        #    3) overwrite the number of data points in the old file last header packet with true value
                        prior_file = open(prior_file_name, 'rb+')
                        if file_num < 10:          numstr = "_00" + str(file_num)
                        elif 10 <= file_num < 100: numstr = "_0" + str(file_num)
                        else:                      numstr = "_" + str(file_num)
                        subset_file = open(file_name + numstr + file_ext, 'wb')
                        print(("Writing subset file: " + ospath.split(subset_file.name)[1]))

                        if self.basic_header['FileSpec']  == '2.1':
                            subset_file.write(prior_file.read(32 + 4 * num_elecs))
                        else:
                            subset_file.write(prior_file.read(bytes_in_headers))
                            subset_file.write(header_binary)
                            timestamp_new = unpack('<I', timestamp_binary)[0] \
                                            + (packet_read_pts + pts_can_add) * self.basic_header['Period']
                            subset_file.write(np.array(timestamp_new).astype(np.uint32).tobytes())
                            subset_file.write(np.array(num_pts - pts_can_add).astype(np.uint32).tobytes())

                            prior_file.seek(num_pts_header_pos, 0)
                            prior_file.write(np.array(prior_file_pkt_pts).astype(np.uint32).tobytes())

                            num_pts_header_pos = bytes_in_headers + 5

                        # Close old file and update parameters
                        prior_file.close()
                        packet_read_pts     += pts_can_add
                        start_idx           += pts_can_add
                        num_pts             -= pts_can_add
                        file_num            += 1
                        subset_file_pkt_pts  = 0
                        pausing              = False

                        pts_can_add = int((file_size - subset_file.tell()) // (num_elecs * DATA_BYTE_SIZE)) + 1
                        stop_idx    = start_idx + pts_can_add

                # If no additional file needed, write remaining data to disk, update parameters, and clear memory map
                if elec_id_indices:  subset_file.write(np.array(mm[start_idx:]).tobytes())
                else:                subset_file.write(mm[start_idx:])
                packet_read_pts     += num_pts
                subset_file_pkt_pts += num_pts
                del mm

            # Update num_pts header position for each packet, while saving last packet num_pts_header_pos for later
            if self.basic_header['FileSpec'] != '2.1':
                curr_hdr_num_pts_pos = num_pts_header_pos
                num_pts_header_pos  += 4 + subset_file_pkt_pts * num_elecs * DATA_BYTE_SIZE + 5

            # Because memory map resets the file position, reset position in datafile
            datafile_pos += self.basic_header['ChannelCount'] * packet_pts * DATA_BYTE_SIZE
            self.datafile.seek(datafile_pos, 0)

            # If using file_timing and there is pausing in data (multiple packets), let user know
            if file_time_s and not pausing and (self.datafile.tell() != ospath.getsize(self.datafile.name)):
                pausing = True
                print("\n*** Because of pausing in original datafile, this file may be slightly time shorter\n"
                      "       than others, and will contain multiple data packets offset in time\n")

            # Update last data header packet num data points accordingly (spec != 2.1)
            if self.basic_header['FileSpec'] != '2.1':
                subset_file_pos = subset_file.tell()
                subset_file.seek(curr_hdr_num_pts_pos, 0)
                subset_file.write(np.array(subset_file_pkt_pts).astype(np.uint32).tobytes())
                subset_file.seek(subset_file_pos, 0)

        # Close subset file and return success
        subset_file.close()
        print("\n *** All subset files written to disk and closed ***")
        return "SUCCESS"

    def close(self):
        name = self.datafile.name
        self.datafile.close()
        print(('\n' + name.split('/')[-1] + ' closed'))



from riglib.bmi import lindecoder, kfdecoder, state_space_models, extractor, train
from built_in_tasks.bmimultitasks import BMIControlMulti #, SimBMICosEncLinDec, SimBMIVelocityLinDec
from features.ecube_features import EcubeBMI, EcubeFileBMI, RecordECube
from riglib.stereo_opengl.window import WindowDispl2D
from riglib import experiment
from features.hdf_features import SaveHDF
import numpy as np

import unittest


# might need this on windows:
# export LIBGL_ALWAYS_INDIRECT=0
#export DISPLAY=$(grep nameserver /etc/resolv.conf | awk '{print $2}'):0.0

class TestKFDecoder(unittest.TestCase):
        
    @unittest.skip("")
    def test_fixed_decoder_ecube(self):

         # Construct a fixed decoder
        ssm = state_space_models.StateSpaceEndptVel2D()
        units = np.array([[1, 0], [2, 0]])
        C = np.zeros([2, 7])
        C[0, 3] = 0.1
        C[1, 5] = 0.1
        decoder = train.make_fixed_kf_decoder(units, ssm, C, dt=0.1)
        decoder.extractor_cls = extractor.LFPMTMPowerExtractor
        decoder.extractor_kwargs = dict(channels=[1, 2], bands=[(90,110)], win_len=0.1, fs=1000)

        import pickle
        import os
        test_decoder_filename = os.path.join('tests', 'test_kf_decoder.pkl')
        with open(test_decoder_filename, 'wb') as f:
            pickle.dump(decoder, f, 2)

        del decoder
        
        with open(test_decoder_filename, 'rb') as f:
            decoder = pickle.load(f)

        base_class = BMIControlMulti

        # Settings for streaming from ecube
        # feats = [EcubeBMI, WindowDispl2D, SaveHDF] # use default headstage port 7
        # kwargs = dict(decoder=decoder, window_size=(500,500), fullscreen=False)

        # Settings for reading from file
        feats = [EcubeFileBMI, WindowDispl2D, SaveHDF]
        test_file = 'tests/test_data/simple'
        # test_file = '/media/server/raw/ecube/ecube test data'
        kwargs = dict(ecube_bmi_filename=test_file, decoder=decoder)

        seq = base_class.centerout_2D(nblocks=1, ntargets=8, distance=8)
        Exp = experiment.make(base_class, feats=feats)
        exp = Exp(seq, **kwargs)

        exp.window_size = (500,500)
        exp.fullscreen = False
        exp.init()

        print(f"decoder units: {exp.decoder.units}")
        print(f"decoder binlen: {exp.decoder.binlen}")
        print(f"decoder call rate: {exp.decoder.call_rate}")

        exp.run()

        h5file = exp.get_h5_filename()
        os.rename(h5file, 'test_decoder.hdf')
        
        rewards, time_penalties, hold_penalties = calculate_rewards(exp)
        self.assertTrue(rewards <= rewards + time_penalties + hold_penalties)
        self.assertTrue(rewards > 0)

    @unittest.skip('msg')
    def test_trained_decoder_simulation(self):
        import aopy

        # Train a decoder form test neural and cursor data generated from a known encoder model
        data = aopy.data.load_hdf_group('tests/test_data/feature_selection', 'wo_FS_0.7_training_data.hdf')
        position = data['kinematics']
        velocity = np.diff(position.T, axis=0) * 1./(1/60)
        velocity = np.vstack([np.zeros(position.shape[0]), velocity])
        kin = np.hstack([position.T, velocity])
        print(kin.shape)
        units = np.array([[i+1, 0] for i in range(8)])
        neural_features = data['spike_counts']
        print(neural_features.shape)
        update_rate = 60
        ssm = state_space_models.StateSpaceEndptVel2D()
        decoder = train.train_KFDecoder_abstract(ssm, kin.T, neural_features.T, units, update_rate)

        print(np.round(decoder.kf.C, 3))
        print(np.round(decoder.kf.Q, 3))

    def test_trained_decoder_ecube(self):
        import aopy

        # Train a decoder form test neural and cursor data generated from a known encoder model
        data = aopy.data.load_hdf_group('tests/test_data/feature_selection', 'wo_FS_0.7_training_data.hdf')
        winlen = 0.5
        position = data['kinematics'][:,::int(winlen*60)] # only need one kin sample per extractor window size
        velocity = np.diff(position.T, axis=0) * 1./(1/60)
        velocity = np.vstack([np.zeros(position.shape[0]), velocity])
        kin = np.hstack([position.T, velocity])
        print(kin.shape)
        files = {'ecube': 'tests/test_data/feature_selection'}
        units = np.array([[i+1, 0] for i in range(8)])
        print(data['neurows'].shape)
        print(data['neurows'])
        neural_features, units, extractor_kwargs = extractor.LFPMTMPowerExtractor.extract_from_file(files, data['neurows'], winlen, units, {'channels': [i+1 for i in range(8)], 'bands': [(70,90)], 'win_len': winlen})
        print(neural_features.shape)
        
        import matplotlib.pyplot as plt
        plt.plot(neural_features[:,2])
        plt.plot(kin[:,2])
        plt.show()

        update_rate = 60
        ssm = state_space_models.StateSpaceEndptVel2D()
        decoder = train.train_KFDecoder_abstract(ssm, kin.T, neural_features.T, units, update_rate, zscore=True)
        decoder.extractor_cls = extractor.LFPMTMPowerExtractor
        decoder.extractor_kwargs = extractor_kwargs
        print(np.round(decoder.kf.C, 3))
        print(np.round(decoder.kf.Q, 3))

        # Save the decoder
        import pickle
        import os
        test_decoder_filename = os.path.join('tests', 'trained_kf_decoder.pkl')
        with open(test_decoder_filename, 'wb') as f:
            pickle.dump(decoder, f, 2)

        # Load the sequence
        data = aopy.data.load_hdf_group('tests/test_data/feature_selection', 'wo_FS_0.7_training_data.hdf')
        targ_seq = data['target_sequence']
        targ_locs = data['target_location']
        seq = list(zip([[i] for i in targ_seq], [[l] for l in targ_locs]))

        # Make and run the experiment
        base_class = BMIControlMulti
        #feats = [EcubeBMI] # use default headstage port 7
        feats = [EcubeFileBMI, WindowDispl2D]
        kwargs = dict(ecube_bmi_filename='tests/test_data/feature_selection', decoder=decoder)
        Exp = experiment.make(base_class, feats=feats)
        exp = Exp(seq, **kwargs)

        exp.window_size = (500,500)
        exp.fullscreen = False

        print(exp.decoder.units)
        print(exp.decoder.units[:,0])
        print(exp.cortical_channels)

        exp.init()
        exp.run()

        # Do some simple checks
        rewards, time_penalties, hold_penalties = calculate_rewards(exp)
        self.assertTrue(rewards <= rewards + time_penalties + hold_penalties)
        self.assertTrue(rewards > 0)


class TestLinDec(unittest.TestCase):

    def test_sanity(self):
        simple_filt = lindecoder.LinearScaleFilter(100, 1, 1)
        self.assertEqual(0, simple_filt.get_mean())
        
        for i in range(50):
            simple_filt([1])

        self.assertEqual(0.5, np.mean(simple_filt.obs))
        self.assertEqual(0, simple_filt.get_mean())

        for i in range(250):
            simple_filt(i)

        self.assertTrue(simple_filt.get_mean() > 0)

    def test_filter(self):
        filt = lindecoder.LinearScaleFilter(100, 3, 2)
        self.assertListEqual([0,0,0], filt.get_mean().tolist())
        for i in range(100):
            filt([0, 0])
            self.assertEqual(0, filt.state.mean[0, 0])
            self.assertEqual(0, filt.state.mean[1, 0])
            self.assertEqual(0, filt.state.mean[2, 0])
    
    @unittest.skip('msg')
    def test_experiment_unfixed(self):
        from built_in_tasks.bmimultitasks import SimBMICosEncLinDec, SimBMIVelocityLinDec
        for cls in [SimBMICosEncLinDec]:
            N_TARGETS = 8
            N_TRIALS = 16
            seq = cls.sim_target_no_center(
                N_TARGETS, N_TRIALS)
            base_class = cls
            feats = []
            Exp = experiment.make(base_class, feats=feats)
            exp = Exp(seq)
            exp.init()

            exp.run()
            
            rewards, time_penalties, hold_penalties = calculate_rewards(exp)
            self.assertTrue(rewards <= rewards + time_penalties + hold_penalties)
            self.assertTrue(rewards > 0)
    
    @unittest.skip('msg')
    def test_experiment(self):
        from built_in_tasks.bmimultitasks import SimBMICosEncLinDec, SimBMIVelocityLinDec
        for cls in [SimBMICosEncLinDec, SimBMIVelocityLinDec]:
            N_TARGETS = 8
            N_TRIALS = 16
            seq = cls.sim_target_seq_generator_multi(
                N_TARGETS, N_TRIALS)
            base_class = cls
            feats = []
            Exp = experiment.make(base_class, feats=feats)
            exp = Exp(seq)
            exp.init()
            exp.decoder.filt.fix_norm_attr()

            exp.run()
            
            rewards, time_penalties, hold_penalties = calculate_rewards(exp)
            self.assertTrue(rewards <= rewards + time_penalties + hold_penalties)
            self.assertTrue(rewards > 0)



def calculate_rewards(exp):
    rewards = 0
    time_penalties = 0
    hold_penalties = 0
    for s in exp.event_log:
        if s[0] == 'reward':
            rewards += 1
        elif s[0] == 'hold_penalty':
            hold_penalties += 1
        elif s[0] == 'timeout_penalty':
            time_penalties += 1
    return rewards, time_penalties, hold_penalties


if __name__ == '__main__':
    unittest.main()





import sys

import time
import numpy as np

import os

from riglib.pupillabs import utils
from riglib.stereo_opengl.textures import Texture
os.environ['DISPLAY'] = ':0'

from riglib.stereo_opengl.environment import Grid
from riglib.stereo_opengl.window import Window, Window2D, FPScontrol
from riglib.stereo_opengl.primitives import AprilTag, Cylinder, Cube, Plane, Sphere, Cone, Text, TexSphere, TexCube, TexPlane
from features.optitrack_features import SpheresToCylinders
from riglib.stereo_opengl.window import Window, Window2D, FPScontrol, WindowSSAO
from riglib.stereo_opengl.openxr import WindowVR
from riglib.stereo_opengl.environment import Box, Grid
from riglib.stereo_opengl.primitives import Cylinder, Cube, Plane, Sphere, Cone, Text, TexSphere, TexCube, TexPlane
from riglib.stereo_opengl.models import FlatMesh, Group
from riglib.stereo_opengl.render import ssao, stereo, Renderer
from riglib.stereo_opengl.utils import cloudy_tex

from riglib.stereo_opengl.ik import RobotArmGen2D
from riglib.stereo_opengl.xfm import Quaternion, Transform
import time

from riglib.stereo_opengl.ik import RobotArm

from built_in_tasks.target_graphics import TextTarget, target_colors
from built_in_tasks.target_capture_task import ScreenTargetCapture
import pygame
from OpenGL.GL import *

# arm4j = RobotArmGen2D(link_radii=.2, joint_radii=.2, link_lengths=[4,4,2,2])
travel_speed = 0.5
travel_radius = 10
moon = Sphere(radius=0.5, color=[0.25,0.25,0.75,0.5])
planet = Sphere(3, color=[0.75,0.25,0.25,0.75])
orbit_radius = 4
orbit_speed = 1
wobble_radius = 0
wobble_speed = 0.5
#TexSphere = type("TexSphere", (Sphere, TexModel), {})
#TexPlane = type("TexPlane", (Plane, TexModel), {})
#reward_text = Text(7.5, "123", justify='right', color=[1,0,1,1])
# center_out_gen = ScreenTargetCapture.centerout_2D(1)
# center_out_positions = [pos[1] for _, pos in center_out_gen]
center_out_gen = ScreenTargetCapture.centerout_tabletop(1)
center_out_positions = [(pos[1][0], pos[1][1], -10) for _, pos in center_out_gen]
center_out_targets = [
    Sphere(radius=2, color=target_colors['yellow']).translate(*pos)
    for pos in center_out_positions
]
center_out_targets[6].color = target_colors['cyan']

pos_list = np.array([[0,0,0],[0,0,5]])

class Test2(Window):

    def __init__(self, *args, **kwargs):
        self.count=0
        self.cursor_bounds = np.array([-15, 15, -15, 15, -15, 15])
        super().__init__(*args, **kwargs)

    def _start_draw(self):
        #arm4j.set_joint_pos([0,0,np.pi/2,np.pi/2])
        #arm4j.get_endpoint_pos()
        self.add_model(Grid(50))
        # self.add_model(moon)
        # self.add_model(planet)
        # self.add_model(arm4j)
        #self.add_model(reward_text.translate(5,0,-5))
        # self.add_model(TexSphere(radius=3, specular_color=[1,1,1,1], tex=cloudy_tex()).translate(5,0,0))
        # self.add_model(TexPlane(5,5, tex=cloudy_tex(), specular_color=(0.,0,0,1)).rotate_x(90))
        # self.add_model(TexPlane(5,5, specular_color=(0.,0,0,1), tex=cloudy_tex()).rotate_x(90))
        # reward_text = Text(7.5, "123", justify='right', color=[1,0,1,0.75])
        # self.add_model(reward_text)
        # self.add_model(TexPlane(4,4,color=[0,0,0,0.9], tex=cloudy_tex()).rotate_x(90).translate(0,0,-5))
        #self.screen_init()
        #self.draw_world()
        for model in center_out_targets:
            self.add_model(model)
        self.add_model(Sphere(radius=1, color=target_colors['purple']).translate(3,3,-10))

    def _while_draw(self):
        ts = time.time() - self.start_time
        
        x = travel_radius * np.cos(ts * travel_speed)
        y = travel_radius * np.sin(ts * travel_speed)

        pos = np.array([x, 0, y])

        x = orbit_radius * np.cos(ts * orbit_speed)
        z = orbit_radius * np.sin(ts * orbit_speed)

        moon.translate(x+pos[0],z+pos[1],pos[2],reset=True)
        planet.translate(pos[0], pos[1], pos[2],reset=True)

        x = wobble_radius * np.cos(ts * wobble_speed)
        y = wobble_radius * np.sin(ts * wobble_speed)

        xfm = Transform(move=[x,0,-self.screen_dist])
        xfm.rotate_x(np.radians(y))
        xfm.rotate_y(np.radians(x))
        self.modelview = xfm.to_mat()

        if ts > 2 and self.count<len(pos_list):
            # reward_text.translate(*pos_list[self.count])
            self.count+=1
        if ts > 4 and self.count<len(pos_list)+1:
            # win.remove_model(reward_text)
            # target = TextTarget('hi', [1,1,0,1], 1)
            # win.add_model(target.model)
            self.count += 1
        self.draw_world()

        error = glGetError()
        if error != GL_NO_ERROR:
            print(f"OpenGL error after drawing: {error}")

if __name__ == "__main__":
    win = Test2(window_size=(1000, 800), fullscreen=False, stereo_mode='projection',
                screen_dist=50, screen_half_height=22.5)
    win.run()



'''
Classes implementing various kinematic chains. This module is perhaps mis-located
as it does not have a direct BMI role but rather contains code which is useful in
supporting BMI control of kinematic chains.

This code depends on the 'robot' module (https://github.com/sgowda/robotics_toolbox)
'''
import numpy as np
try:
    import robot
except ImportError as e:
    import warnings
    warnings.warn("The 'robot' module cannot be found! See https://github.com/sgowda/robotics_toolbox")
    print(e)

import matplotlib.pyplot as plt
import time

pi = np.pi

class KinematicChain(object):
    '''
    Arbitrary kinematic chain (i.e. spherical joint at the beginning of 
    each joint)
    '''
    def __init__(self, link_lengths=[10., 10.], name='', base_loc=np.array([0., 0., 0.]), rotation_convention=1):
        '''
        Docstring 

        Parameters
        ----------
        link_lengths: iterable
            Lengths of all the distances between joints
        base_loc: np.array of shape (3,), default=np.array([0, 0, 0])
            Location of the base of the kinematic chain in an "absolute" reference frame
        '''
        self.n_links = len(link_lengths)
        self.link_lengths = link_lengths
        self.base_loc = base_loc

        assert rotation_convention in [-1, 1]
        self.rotation_convention = rotation_convention

        # Create the robot object. Override for child classes with different types of joints
        self._init_serial_link()
        self.robot.name = name

    def _init_serial_link(self):
        links = []
        for link_length in self.link_lengths:
            link1 = robot.Link(alpha=-pi/2)
            link2 = robot.Link(alpha=pi/2)
            link3 = robot.Link(d=-link_length)
            links += [link1, link2, link3]
        
        # By convention, we start the arm in the XY-plane
        links[1].offset = -pi/2 

        self.robot = robot.SerialLink(links)        

    def calc_full_joint_angles(self, joint_angles):
        '''
        Override in child classes to perform static transforms on joint angle inputs. If some 
        joints are always static (e.g., if the chain only operates in a plane)
        this can avoid unclutter joint angle specifications.
        '''
        return self.rotation_convention * joint_angles

    def full_angles_to_subset(self, joint_angles):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        return joint_angles

    def plot(self, joint_angles):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''

        joint_angles = self.calc_full_joint_angles(joint_angles)
        self.robot.plot(joint_angles)

    def forward_kinematics(self, joint_angles, **kwargs):
        '''
        Calculate forward kinematics using D-H parameter convention

        Parameters
        ----------
        
        Returns
        -------        
        '''
        joint_angles = self.calc_full_joint_angles(joint_angles)
        t, allt = self.robot.fkine(joint_angles, **kwargs)
        self.joint_angles = joint_angles
        self.t = t
        self.allt = allt
        return t, allt

    def apply_joint_limits(self, joint_angles):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        return joint_angles

    def inverse_kinematics(self, target_pos, q_start=None, method='pso', **kwargs):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        if q_start == None:
            q_start = self.random_sample()
        return self.inverse_kinematics_pso(target_pos, q_start, **kwargs)
        # ik_method = getattr(self, 'inverse_kinematics_%s' % method)
        # return ik_method(q_start, target_pos)

    def inverse_kinematics_grad_descent(self, target_pos, starting_config, n_iter=1000, verbose=False, eps=0.01, return_path=False):
        '''
        Default inverse kinematics method is RRT since for redundant 
        kinematic chains, an infinite number of inverse kinematics solutions 
        exist

        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------        
        '''

        q = starting_config
        start_time = time.time()
        endpoint_traj = np.zeros([n_iter, 3])

        joint_limited = np.zeros(len(q))

        for k in range(n_iter):
            # print k
            # calc endpoint position of the manipulator
            endpoint_traj[k] = self.endpoint_pos(q)

            current_cost = np.linalg.norm(endpoint_traj[k] - target_pos, 2)
            if current_cost < eps:
                print("Terminating early")
                break

            # calculate the jacobian
            J = self.jacobian(q)
            J_pos = J[0:3,:]

            # for joints that are at their limit, zero out the jacobian?
            # J_pos[:, np.nonzero(self.calc_full_joint_angles(joint_limited))] = 0

            # take a step from the current position toward the target pos using the inverse Jacobian
            J_inv = np.linalg.pinv(J_pos)
            # J_inv = J_pos.T

            xdot = (target_pos - endpoint_traj[k])#/np.linalg.norm(endpoint_traj[k] - target_pos) 

            # if current_cost < 3 or k > 10:
            #     stepsize = 0.001
            # else:
            #     stepsize = 0.01


            xdot = (target_pos - endpoint_traj[k])#/np.linalg.norm(endpoint_traj[k] - target_pos)
            # xdot = (endpoint_traj[k] - target_pos)/np.linalg.norm(endpoint_traj[k] - target_pos)
            qdot = 0.001*np.dot(J_inv, xdot)
            qdot = self.full_angles_to_subset(np.array(qdot).ravel())

            q += qdot

            # apply joint limits
            q, joint_limited = self.apply_joint_limits(q)

        end_time = time.time()
        runtime = end_time - start_time
        if verbose:
            print("Runtime: %g" % runtime)
            print("# of iterations: %g" % k)

        if return_path:
            return q, endpoint_traj[:k]
        else:
            return q

    def jacobian(self, joint_angles):
        '''
        Return the full jacobian 

        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------        
        '''
        joint_angles = self.calc_full_joint_angles(joint_angles)
        J = self.robot.jacobn(joint_angles)
        return J

    def endpoint_pos(self, joint_angles):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''

        t, allt = self.forward_kinematics(joint_angles)
        pos_rel_to_base = np.array(t[0:3,-1]).ravel()
        return pos_rel_to_base + self.base_loc

    def ik_cost(self, q, q_start, target_pos, weight=100):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''

        q_diff = q - q_start
        return np.linalg.norm(q_diff[0:2]) + weight*np.linalg.norm(self.endpoint_pos(q) - target_pos)

    def inverse_kinematics_pso(self, target_pos, q_start, time_limit=np.inf, verbose=False, eps=0.5, n_particles=10, n_iter=10):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''

        # Initialize the particles; 
        n_joints = self.n_joints

        particles_q = np.tile(q_start, [n_particles, 1])

        # if 0:
        #     # initialize the velocities to be biased around the direction the jacobian tells you is correct
        #     current_pos = self.endpoint_pos(q_start)
        #     int_displ = target_pos - current_pos
        #     print int_displ, target_pos
        #     J = self.jacobian(q_start)
        #     endpoint_vel = np.random.randn(n_particles, 3)# + int_displ
        #     particles_v = np.dot(J[0:3,1::3].T, endpoint_vel.T).T
        # else:
        #     # initialize particle velocities randomly

        
        particles_v = np.random.randn(n_particles, n_joints) #/ np.array([1., 1., 1, 1]) #np.array(self.link_lengths)

        cost_fn = lambda q: self.ik_cost(q, q_start, target_pos)

        gbest = particles_q.copy()
        gbestcost = np.array(list(map(cost_fn, gbest)))
        pbest = gbest[np.argmin(gbestcost)]
        pbestcost = cost_fn(pbest)

        min_limits = np.array([x[0] for x in self.joint_limits])
        max_limits = np.array([x[1] for x in self.joint_limits])
        min_limits = np.tile(min_limits, [n_particles, 1])
        max_limits = np.tile(max_limits, [n_particles, 1])

        start_time = time.time()
        for k in range(n_iter):
            if time.time() - start_time > time_limit:
                break

            # update positions of particles
            particles_q += particles_v

            # apply joint limits
            min_viol = particles_q < min_limits
            max_viol = particles_q > max_limits
            particles_q[min_viol] = min_limits[min_viol]
            particles_q[max_viol] = max_limits[max_viol]

            # update the costs
            costs = np.array(list(map(cost_fn, particles_q)))

            # update the 'bests'
            gbest[gbestcost > costs] = particles_q[gbestcost > costs]
            gbestcost[gbestcost > costs] = costs[gbestcost > costs]

            idx = np.argmin(gbestcost)
            pbest = gbest[idx]
            pbestcost = gbestcost[idx]

            # update the velocity
            phi1 = 1#np.random.rand()
            phi2 = 1#np.random.rand()
            w=0.25
            c1=0.5
            c2=0.25
            particles_v = w*particles_v + c1*phi1*(pbest - particles_q) + c2*phi2*(gbest - particles_q)

            error = np.linalg.norm(self.endpoint_pos(pbest) - target_pos)
            if error < eps:
                break
            
        end_time = time.time()
        if verbose: print("Runtime = %g, error = %g, n_iter=%d" % (end_time-start_time, error, k))

        return pbest

    def spatial_positions_of_joints(self, joint_angles):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''

        _, allt = self.forward_kinematics(joint_angles, return_allt=True)
        pos = (allt[0:3, -1,:].T + self.base_loc).T
        # pos = np.hstack([np.zeros([3,1]), pos])
        return pos


class PlanarXZKinematicChain(KinematicChain):
    '''
    Kinematic chain restricted to movement in the XZ-plane
    '''
    def _init_serial_link(self):
        base = robot.Link(alpha=pi/2, d=0, a=0)
        links = [base]
        for link_length in self.link_lengths:
            link1 = robot.Link(alpha=0, d=0, a=link_length)
            links.append(link1)

            # link2 = robot.Link(alpha=pi/2)
            # link3 = robot.Link(d=-link_length)
            # links += [link1, link2, link3]
        
        # By convention, we start the arm in the XY-plane
        # links[1].offset = -pi/2 

        self.robot = robot.SerialLink(links)

    def calc_full_joint_angles(self, joint_angles):
        '''
        only some joints rotate in the planar kinematic chain

        Parameters
        ----------
        joint_angles : np.ndarray of shape (self.n_links)
            Joint angles without the angle for the base link, which is fixed at 0
        
        Returns
        -------
        joint_angles_full : np.ndarray of shape (self.n_links+1)
            Add on the 0 at the proximal end for the base link angle
        '''
        if not len(joint_angles) == self.n_links:
            raise ValueError("Incorrect number of joint angles specified!")

        # # There are really 3 angles per joint to allow 3D rotation at each joint
        # joint_angles_full = np.zeros(self.n_links * 3)  
        # joint_angles_full[1::3] = joint_angles

        joint_angles_full = np.hstack([0, joint_angles])
        return self.rotation_convention * joint_angles_full 

    def random_sample(self):
        '''
        Sample the joint configuration space within the limits of each joint
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        '''
        if hasattr(self, 'joint_limits'):
            joint_limits = self.joint_limits
        else:
            joint_limits = [(-np.pi, np.pi)] * self.n_links

        q_start = []
        for lim_min, lim_max in joint_limits:
            q_start.append(np.random.uniform(lim_min, lim_max))
        return np.array(q_start)

    def full_angles_to_subset(self, joint_angles):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        # return joint_angles[1::3]
        return joint_angles[1:]

    def apply_joint_limits(self, joint_angles):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''        
        if not hasattr(self, 'joint_limits'):
            return joint_angles
        else:
            angles = []
            limit_hit = []
            for angle, (lim_min, lim_max) in zip(joint_angles, self.joint_limits):
                limit_hit.append(angle < lim_min or angle > lim_max)
                angle = max(lim_min, angle)
                angle = min(angle, lim_max)
                angles.append(angle)

            return np.array(angles), np.array(limit_hit)

    @property 
    def n_joints(self):
        '''
        In a planar arm, the number of joints equals the number of links
        '''
        return len(self.link_lengths)

    def spatial_positions_of_joints(self, *args, **kwargs):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''        
        pos_all_joints = super(PlanarXZKinematicChain, self).spatial_positions_of_joints(*args, **kwargs)
        return pos_all_joints #(pos_all_joints[:,::3].T + self.base_loc).T

    def create_ik_subchains(self):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''        
        proximal_link_lengths = self.link_lengths[:2]
        distal_link_lengths = self.link_lengths[2:]
        self.proximal_chain = PlanarXZKinematicChain2Link(proximal_link_lengths)
        if len(self.link_lengths) > 2:
            self.distal_chain = PlanarXZKinematicChain(distal_link_lengths)
        else:
            self.distal_chain = None

    def inverse_kinematics(self, target_pos, **kwargs):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        target_pos = target_pos.copy()
        target_pos -= self.base_loc
        if not hasattr(self, 'proximal_chain') or not hasattr(self, 'distal_chain'):
            self.create_ik_subchains()

        if len(self.link_lengths) > 2:
            distal_angles = kwargs.pop('distal_angles', None)

            if distal_angles is None:
                # Sample randomly from the joint limits (-pi, pi) if not specified
                if not hasattr(self, 'joint_limits') or len(self.joint_limits) < len(self.link_lengths):
                    joint_limits = [(-pi, pi)] * len(self.distal_chain.link_lengths)
                else:
                    joint_limits = self.joint_limits[2:]
                distal_angles = np.array([np.random.uniform(*limits) for limits in joint_limits])

            distal_displ = self.distal_chain.endpoint_pos(distal_angles)
            proximal_endpoint_pos = target_pos - distal_displ
            proximal_angles = self.proximal_chain.inverse_kinematics(proximal_endpoint_pos).ravel()
            angles = distal_angles.copy()
            joint_angles = proximal_angles.tolist()
            angles[0] -= np.sum(proximal_angles)
            ik_angles = np.hstack([proximal_angles, angles])
            ik_angles = np.array([np.arctan2(np.sin(angle), np.cos(angle)) for angle in ik_angles])
            return ik_angles
        else:
            return self.proximal_chain.inverse_kinematics(target_pos).ravel()

    def jacobian(self, theta, old=False):
        '''
        Returns the first derivative of the forward kinematics function for x and z endpoint positions: 
        [[dx/dtheta_1, ..., dx/dtheta_N]
         [dz/dtheta_1, ..., dz/dtheta_N]]
        
        Parameters
        ----------
        theta : np.ndarray of shape (N,)
            Valid configuration for the arm (the jacobian calculations are specific to the configuration of the arm)
        
        Returns
        -------
        J : np.ndarray of shape (2, N)
            Manipulator jacobian in the format above
        '''   
        if old: 
            # Calculate jacobian based on hand calculation specific to this type of chain
            l = self.link_lengths
            N = len(theta)
            J = np.zeros([2, len(l)])
            for m in range(N):
                for i in range(m, N):
                    J[0, m] += -l[i]*np.sin(sum(self.rotation_convention*theta[:i+1]))
                    J[1, m] +=  l[i]*np.cos(sum(self.rotation_convention*theta[:i+1]))
            return J 
        else:
            # Use the robotics toolbox and the generic D-H convention jacobian
            J = self.robot.jacob0(self.calc_full_joint_angles(theta))
            return np.array(J[[0,2], 1:])

    def endpoint_potent_null_split(self, q, vel, return_J=False):
        '''
        (Approximately) split joint velocities into an endpoint potent component, 
        which moves the endpoint, and an endpoint null component which only causes self-motion
        '''
        J = self.jacobian(q)
        J_pinv = np.linalg.pinv(J)
        J_task = np.dot(J_pinv, J)
        J_null = np.eye(self.n_joints) - J_task

        vel_task = np.dot(J_task, vel)
        vel_null = np.dot(J_null, vel)
        if return_J:
            return vel_task, vel_null, J, J_pinv
        else:
            return vel_task, vel_null

    def config_change_nullspace_workspace(self, config1, config2):
        '''
        For two configurations, determine how much joint displacement is in the "null" space and how much is in the "task" space

        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        config = config1
        vel = config2 - config1
        endpt1 = self.endpoint_pos(config1)
        endpt2 = self.endpoint_pos(config2)
        task_displ = np.linalg.norm(endpt1 - endpt2)
        
        # compute total displ of individual joints
        total_joint_displ = 0
        
        n_joints = len(config1)
        for k in range(n_joints):
            jnt_k_vel = np.zeros(n_joints)
            jnt_k_vel[k] = vel[k]
            single_joint_displ_pos = self.endpoint_pos(config + jnt_k_vel)
            total_joint_displ += np.linalg.norm(endpt1 - single_joint_displ_pos)
            
        return task_displ, total_joint_displ

    def detect_collision(self, theta, obstacle_pos):
        '''
        Detect a collision between the chain and a circular object
        '''
        spatial_joint_pos = self.spatial_positions_of_joints(theta).T + self.base_loc
        plant_segments = [(x, y) for x, y in zip(spatial_joint_pos[:-1], spatial_joint_pos[1:])]
        dist_to_object = np.zeros(len(plant_segments))
        for k, segment in enumerate(plant_segments):
            dist_to_object[k] = point_to_line_segment_distance(obstacle_pos, segment)
        return dist_to_object

    def plot_joint_pos(self, joint_pos, ax=None, flip=False, **kwargs):
        if ax == None:
            plt.figure()
            ax = plt.subplot(111)

        if isinstance(joint_pos, dict):
            joint_pos = np.vstack(list(joint_pos.values()))
        elif isinstance(joint_pos, np.ndarray) and np.ndim(joint_pos) == 1:
            joint_pos = joint_pos.reshape(1, -1)
        elif isinstance(joint_pos, tuple):
            joint_pos = np.array(joint_pos).reshape(1, -1)

        for pos in joint_pos:
            spatial_pos = self.spatial_positions_of_joints(pos).T

            shoulder_anchor = np.array([2., 0., -15.])
            spatial_pos = spatial_pos# + shoulder_anchor
            if flip:
                ax.plot(-spatial_pos[:,0], spatial_pos[:,2], **kwargs)
            else:
                ax.plot(spatial_pos[:,0], spatial_pos[:,2], **kwargs)
        
        return ax

def point_to_line_segment_distance(point, segment):
    '''
    Determine the distance between a point and a line segment. Used to determine collisions between robot arm links and virtual obstacles.
    Adapted from http://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment
    '''
    v, w = segment
    l2 = np.sum(np.abs(v - w)**2)
    if l2 == 0:
        return np.linalg.norm(v - point)

    t = np.dot(point - v, w - v)/l2
    if t < 0:
        return np.linalg.norm(point - v)
    elif t > 1:
        return np.linalg.norm(point - w)
    else:
        projection = v + t*(w-v)
        return np.linalg.norm(projection - point)


class PlanarXZKinematicChain2Link(PlanarXZKinematicChain):
    ''' Docstring '''
    def __init__(self, link_lengths, *args, **kwargs):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        if not len(link_lengths) == 2:
            raise ValueError("Can't instantiate a 2-link arm with > 2 links!")

        super(PlanarXZKinematicChain2Link, self).__init__(link_lengths, *args, **kwargs)

    def inverse_kinematics(self, pos, **kwargs):
        '''
        Inverse kinematics for a two-link kinematic chain. These equations can be solved
        deterministically. 

        Docstring    
        
        Parameters
        ----------
        pos : np.ndarray of shape (3,)
            Desired endpoint position where the coordinate system origin is the base of the arm. y coordinate must be 0
        
        Returns
        -------
        np.ndarray of shape (2,)
            Joint angles which yield the endpoint position with the forward kinematics of this manipulator
        '''
        pos -= self.base_loc        
        l_upperarm, l_forearm = self.link_lengths 

        if np.ndim(pos) == 1:
            pos = pos.reshape(1,-1)

        # require the y-coordinate to be 0, i.e. flat on the screen
        x, y, z = pos[:,0], pos[:,1], pos[:,2]
        assert np.all(np.abs(np.array(y)) < 1e-10)

        L = np.sqrt(x**2 + z**2)
        cos_el_pflex = (L**2 - l_forearm**2 - l_upperarm**2) / (2*l_forearm*l_upperarm)

        cos_el_pflex[ (cos_el_pflex > 1) & (cos_el_pflex < 1 + 1e-9)] = 1
        el_pflex = np.arccos(cos_el_pflex)

        sh_pabd = np.arctan2(z, x) - np.arcsin(l_forearm * np.sin(np.pi - el_pflex) / L)
        return np.array([sh_pabd, el_pflex])